@@ -8,9 +8,6 @@ struct parasite_ctl;
 #define ARCH_SI_TRAP SI_KERNEL
 
 
-extern const char code_syscall[];
-extern const int code_syscall_size;
-
 void parasite_setup_regs(unsigned long new_ip, void *stack, user_regs_struct_t *regs);
 
 void *mmap_seized(struct parasite_ctl *ctl,
