@@ -21,8 +21,9 @@
 #include <sys/prctl.h>
 #include <sched.h>
 #include <linux/aio_abi.h>
+#include <sys/mount.h>
 
-#include "proc_parse.h"
+#include "fdinfo.h"
 #include "sockets.h"
 #include "crtools.h"
 #include "log.h"
@@ -41,6 +42,8 @@
 #include "pstree.h"
 #include "cr_options.h"
 
+static char *feature_name(int (*func)());
+
 static int check_tty(void)
 {
 	int master = -1, slave = -1;
@@ -161,15 +164,15 @@ static int check_kcmp(void)
 {
 	int ret = syscall(SYS_kcmp, getpid(), -1, -1, -1, -1);
 
-	if (ret != -ENOSYS)
-		return 0;
+	if (ret < 0 && errno == ENOSYS) {
+		pr_perror("System call kcmp is not supported");
+		return -1;
+	}
 
-	errno = -ret;
-	pr_perror("System call kcmp is not supported");
-	return -1;
+	return 0;
 }
 
-static int check_prctl(void)
+static int check_prctl_cat1(void)
 {
 	unsigned long user_auxv = 0;
 	unsigned int *tid_addr;
@@ -177,41 +180,36 @@ static int check_prctl(void)
 	int ret;
 
 	ret = prctl(PR_GET_TID_ADDRESS, (unsigned long)&tid_addr, 0, 0, 0);
-	if (ret) {
-		pr_msg("prctl: PR_GET_TID_ADDRESS is not supported");
+	if (ret < 0) {
+		pr_msg("prctl: PR_GET_TID_ADDRESS is not supported: %m");
 		return -1;
 	}
 
 	/*
-	 * Either new or old interface must be supported in the kernel.
+	 * It's OK if the new interface is not supported because it's
+	 * a Category 2 feature, but the old interface has to be supported.
 	 */
 	ret = prctl(PR_SET_MM, PR_SET_MM_MAP_SIZE, (unsigned long)&size, 0, 0);
-	if (ret) {
-		if (!opts.check_ms_kernel) {
-			pr_msg("prctl: PR_SET_MM_MAP is not supported, which "
-			       "is required for restoring user namespaces\n");
-			return -1;
-		} else
-			pr_warn("Skipping unssuported PR_SET_MM_MAP\n");
-
-		ret = prctl(PR_SET_MM, PR_SET_MM_BRK, brk(0), 0, 0);
-		if (ret) {
-			if (ret == -EPERM)
+	if (ret < 0) {
+		pr_msg("Info  prctl: PR_SET_MM_MAP_SIZE is not supported\n");
+		ret = prctl(PR_SET_MM, PR_SET_MM_BRK, (unsigned long)sbrk(0), 0, 0);
+		if (ret < 0) {
+			if (errno == EPERM)
 				pr_msg("prctl: One needs CAP_SYS_RESOURCE capability to perform testing\n");
 			else
-				pr_msg("prctl: PR_SET_MM is not supported\n");
+				pr_msg("prctl: PR_SET_MM_BRK is not supported: %m\n");
 			return -1;
 		}
 
 		ret = prctl(PR_SET_MM, PR_SET_MM_EXE_FILE, -1, 0, 0);
-		if (ret != -EBADF) {
-			pr_msg("prctl: PR_SET_MM_EXE_FILE is not supported (%d)\n", ret);
+		if (ret < 0 && errno != EBADF) {
+			pr_msg("prctl: PR_SET_MM_EXE_FILE is not supported: %m\n");
 			return -1;
 		}
 
 		ret = prctl(PR_SET_MM, PR_SET_MM_AUXV, (long)&user_auxv, sizeof(user_auxv), 0);
-		if (ret) {
-			pr_msg("prctl: PR_SET_MM_AUXV is not supported\n");
+		if (ret < 0) {
+			pr_msg("prctl: PR_SET_MM_AUXV is not supported: %m\n");
 			return -1;
 		}
 	}
@@ -219,6 +217,19 @@ static int check_prctl(void)
 	return 0;
 }
 
+static int check_prctl_cat2(void)
+{
+	unsigned int size = 0;
+	int ret;
+
+	ret = prctl(PR_SET_MM, PR_SET_MM_MAP_SIZE, (unsigned long)&size, 0, 0);
+	if (ret) {
+		pr_warn("prctl: PR_SET_MM_MAP_SIZE is not supported\n");
+		return -1;
+	}
+	return 0;
+}
+
 static int check_fcntl(void)
 {
 	u32 v[2];
@@ -523,7 +534,7 @@ static int check_sigqueuinfo()
 
 	signal(SIGUSR1, SIG_IGN);
 
-	if (syscall(SYS_rt_sigqueueinfo, getpid(), SIGUSR1, &info)) {
+	if (syscall(SYS_rt_sigqueueinfo, getpid(), SIGUSR1, &info) < 0) {
 		pr_perror("Unable to send siginfo with positive si_code to itself");
 		return -1;
 	}
@@ -682,8 +693,10 @@ static int check_mem_dirty_track(void)
 	if (kerndat_get_dirty_track() < 0)
 		return -1;
 
-	if (!kdat.has_dirty_track)
+	if (!kdat.has_dirty_track) {
 		pr_warn("Dirty tracking is OFF. Memory snapshot will not work.\n");
+		return -1;
+	}
 	return 0;
 }
 
@@ -744,7 +757,7 @@ static int check_aio_remap(void)
 	int r;
 
 	if (syscall(SYS_io_setup, 16, &ctx) < 0) {
-		pr_err("No AIO syscall\n");
+		pr_err("No AIO syscall: %m\n");
 		return -1;
 	}
 
@@ -766,11 +779,8 @@ static int check_aio_remap(void)
 	ctx = (aio_context_t)naddr;
 	r = syscall(SYS_io_getevents, ctx, 0, 1, NULL, NULL);
 	if (r < 0) {
-		if (!opts.check_ms_kernel) {
-			pr_err("AIO remap doesn't work properly\n");
-			return -1;
-		} else
-			pr_warn("Skipping unsupported AIO remap\n");
+		pr_err("AIO remap doesn't work properly: %m\n");
+		return -1;
 	}
 
 	return 0;
@@ -782,12 +792,8 @@ static int check_fdinfo_lock(void)
 		return -1;
 
 	if (!kdat.has_fdinfo_lock) {
-		if (!opts.check_ms_kernel) {
-			pr_err("fdinfo doesn't contain the lock field\n");
-			return -1;
-		} else {
-			pr_warn("fdinfo doesn't contain the lock field\n");
-		}
+		pr_err("fdinfo doesn't contain the lock field\n");
+		return -1;
 	}
 
 	return 0;
@@ -820,13 +826,87 @@ static int check_clone_parent_vs_pid()
 	return 0;
 }
 
+static int check_autofs_pipe_ino(void)
+{
+	FILE *f;
+	char str[1024];
+	int ret = -ENOENT;
+
+	f = fopen_proc(PROC_SELF, "mountinfo");
+	if (!f) {
+		pr_perror("Can't open %d mountinfo", getpid());
+		return -1;
+	}
+
+	while (fgets(str, sizeof(str), f)) {
+		if (strstr(str, " autofs ")) {
+			if (strstr(str, "pipe_ino="))
+				ret = 0;
+			else {
+				pr_err("autofs not supported.\n");
+				ret = -ENOTSUP;
+			}
+			break;
+		}
+	}
+
+	fclose(f);
+	return ret;
+}
+
+static int check_autofs(void)
+{
+	char *dir, *options, template[] = "/tmp/.criu.mnt.XXXXXX";
+	int ret, pfd[2];
+
+	ret = check_autofs_pipe_ino();
+	if (ret != -ENOENT)
+		return ret;
+
+	if (pipe(pfd) < 0) {
+		pr_perror("failed to create pipe");
+		return -1;
+	}
+
+	ret = -1;
+
+	options = xsprintf("fd=%d,pgrp=%d,minproto=5,maxproto=5,direct",
+				pfd[1], getpgrp());
+	if (!options) {
+		pr_err("failed to allocate autofs options\n");
+		goto close_pipe;
+	}
+
+	dir = mkdtemp(template);
+	if (!dir) {
+		pr_perror("failed to construct temporary name");
+		goto free_options;
+	}
+
+	if (mount("criu", dir, "autofs", 0, options) < 0) {
+		pr_perror("failed to mount autofs");
+		goto unlink_dir;
+	}
+
+	ret = check_autofs_pipe_ino();
+
+	if (umount(dir))
+		pr_perror("failed to umount %s", dir);
+
+unlink_dir:
+	if (rmdir(dir))
+		pr_perror("failed to unlink %s", dir);
+free_options:
+	free(options);
+close_pipe:
+	close(pfd[0]);
+	close(pfd[1]);
+	return ret;
+}
+
 static int check_cgroupns(void)
 {
 	int ret;
-	if (opts.check_ms_kernel) {
-		pr_warn("Skipping cgroup namespaces check\n");
-		return 0;
-	}
 
 	ret = access("/proc/self/ns/cgroup", F_OK);
 	if (ret < 0) {
@@ -837,8 +917,47 @@ static int check_cgroupns(void)
 	return 0;
 }
 
+static int check_tcp_window(void)
+{
+	int ret;
+
+	ret = kerndat_tcp_repair_window();
+	if (ret < 0)
+		return -1;
+
+	if (!kdat.has_tcp_window) {
+		pr_err("The TCP_REPAIR_WINDOW option isn't supported.\n");
+		return -1;
+	}
+
+	return 0;
+}
+
 static int (*chk_feature)(void);
 
+/*
+ * There are three categories of kernel features:
+ *
+ * 1. Absolutely required (/proc/pid/map_files, ptrace PEEKSIGINFO, etc.).
+ * 2. Required only for specific cases (aio remap, tun, etc.).
+ *    Checked when --extra or --all is specified.
+ * 3. Experimental (task-diag).
+ *    Checked when --experimental or --all is specified.
+ *
+ * We fail if any feature in category 1 is missing but tolerate failures
+ * in the other categories.  Currently, there is nothing in category 3.
+ */
+#define CHECK_GOOD	"Looks good."
+#define CHECK_BAD	"Does not look good."
+#define CHECK_MAYBE	"Looks good but some kernel features are missing\n" \
+			"which, depending on your process tree, may cause\n" \
+			"dump or restore failure."
+#define CHECK_CAT1(fn)	do { \
+				if ((ret = fn) != 0) { \
+					print_on_level(DEFAULT_LOGLEVEL, "%s\n", CHECK_BAD); \
+					return ret; \
+				} \
+			} while (0)
 int cr_check(void)
 {
 	struct ns_id ns = { .type = NS_CRIU, .ns_pid = PROC_SELF, .nd = &mnt_ns_desc };
@@ -863,44 +982,71 @@ int cr_check(void)
 		return -1;
 
 	if (chk_feature) {
-		ret = chk_feature();
-		goto out;
+		if (chk_feature())
+			return -1;
+		print_on_level(DEFAULT_LOGLEVEL, "%s is supported\n",
+			feature_name(chk_feature));
+		return 0;
 	}
 
-	ret |= check_map_files();
-	ret |= check_sock_diag();
-	ret |= check_ns_last_pid();
-	ret |= check_sock_peek_off();
-	ret |= check_kcmp();
-	ret |= check_prctl();
-	ret |= check_fcntl();
-	ret |= check_proc_stat();
-	ret |= check_tcp();
-	ret |= check_fdinfo_ext();
-	ret |= check_unaligned_vmsplice();
-	ret |= check_tty();
-	ret |= check_so_gets();
-	ret |= check_ipc();
-	ret |= check_sigqueuinfo();
-	ret |= check_ptrace_peeksiginfo();
-	ret |= check_ptrace_suspend_seccomp();
-	ret |= check_ptrace_dump_seccomp_filters();
-	ret |= check_mem_dirty_track();
-	ret |= check_posix_timers();
-	ret |= check_tun_cr(0);
-	ret |= check_timerfd();
-	ret |= check_mnt_id();
-	ret |= check_aio_remap();
-	ret |= check_fdinfo_lock();
-	ret |= check_clone_parent_vs_pid();
-	ret |= check_cgroupns();
+	/*
+	 * Category 1 - absolutely required.
+	 * So that the user can see clearly what's missing, we exit with
+	 * non-zero status on the first failure because it gets very
+	 * confusing when there are many warnings and error messages.
+	 */
+	CHECK_CAT1(check_map_files());
+	CHECK_CAT1(check_sock_diag());
+	CHECK_CAT1(check_ns_last_pid());
+	CHECK_CAT1(check_sock_peek_off());
+	CHECK_CAT1(check_kcmp());
+	CHECK_CAT1(check_prctl_cat1());
+	CHECK_CAT1(check_fcntl());
+	CHECK_CAT1(check_proc_stat());
+	CHECK_CAT1(check_tcp());
+	CHECK_CAT1(check_fdinfo_ext());
+	CHECK_CAT1(check_unaligned_vmsplice());
+	CHECK_CAT1(check_tty());
+	CHECK_CAT1(check_so_gets());
+	CHECK_CAT1(check_ipc());
+	CHECK_CAT1(check_sigqueuinfo());
+	CHECK_CAT1(check_ptrace_peeksiginfo());
 
-out:
-	if (!ret)
-		print_on_level(DEFAULT_LOGLEVEL, "Looks good.\n");
+	/*
+	 * Category 2 - required for specific cases.
+	 * Unlike Category 1 features, we don't exit with non-zero status
+	 * on a failure because CRIU may still work.
+	 */
+	if (opts.check_extra_features) {
+		ret |= check_prctl_cat2();
+		ret |= check_ptrace_suspend_seccomp();
+		ret |= check_ptrace_dump_seccomp_filters();
+		ret |= check_mem_dirty_track();
+		ret |= check_posix_timers();
+		ret |= check_tun_cr(0);
+		ret |= check_timerfd();
+		ret |= check_mnt_id();
+		ret |= check_aio_remap();
+		ret |= check_fdinfo_lock();
+		ret |= check_clone_parent_vs_pid();
+		ret |= check_cgroupns();
+		ret |= check_tcp_window();
+	}
 
+	/*
+	 * Category 3 - experimental.
+	 */
+	if (opts.check_experimental_features) {
+		ret |= check_autofs();
+	}
+
+	print_on_level(DEFAULT_LOGLEVEL, "%s\n", ret ? CHECK_MAYBE : CHECK_GOOD);
 	return ret;
 }
+#undef CHECK_GOOD
+#undef CHECK_BAD
+#undef CHECK_MAYBE
+#undef CHECK_CAT1
 
 static int check_tun(void)
 {
@@ -925,9 +1071,8 @@ static int check_userns(void)
 	}
 
 	ret = prctl(PR_SET_MM, PR_SET_MM_MAP_SIZE, (unsigned long)&size, 0, 0);
-	if (ret) {
-		errno = -ret;
-		pr_perror("No new prctl API");
+	if (ret < 0) {
+		pr_perror("prctl: PR_SET_MM_MAP_SIZE is not supported");
 		return -1;
 	}
 
@@ -947,32 +1092,72 @@ static int check_loginuid(void)
 	return 0;
 }
 
+struct feature_list {
+	char *name;
+	int (*func)();
+};
+
+static struct feature_list feature_list[] = {
+	{ "mnt_id", check_mnt_id },
+	{ "mem_dirty_track", check_mem_dirty_track },
+	{ "aio_remap", check_aio_remap },
+	{ "timerfd", check_timerfd },
+	{ "tun", check_tun },
+	{ "userns", check_userns },
+	{ "fdinfo_lock", check_fdinfo_lock },
+	{ "seccomp_suspend", check_ptrace_suspend_seccomp },
+	{ "seccomp_filters", check_ptrace_dump_seccomp_filters },
+	{ "loginuid", check_loginuid },
+	{ "cgroupns", check_cgroupns },
+	{ "autofs", check_autofs },
+	{ NULL, NULL },
+};
+
+void pr_check_features(const char *offset, const char *sep, int width)
+{
+	struct feature_list *fl;
+	int pos = width + 1;
+	int sep_len = strlen(sep);
+	int offset_len = strlen(offset);
+
+	for (fl = feature_list; fl->name; fl++) {
+		int len = strlen(fl->name);
+
+		if (pos + len + sep_len > width) {
+			pr_msg("\n%s", offset);
+			pos = offset_len;
+		}
+		pr_msg("%s", fl->name);
+		pos += len;
+		if ((fl + 1)->name) { // not the last item
+			pr_msg("%s", sep);
+			pos += sep_len;
+		}
+	}
+	pr_msg("\n");
+}
+
 int check_add_feature(char *feat)
 {
-	if (!strcmp(feat, "mnt_id"))
-		chk_feature = check_mnt_id;
-	else if (!strcmp(feat, "aio_remap"))
-		chk_feature = check_aio_remap;
-	else if (!strcmp(feat, "timerfd"))
-		chk_feature = check_timerfd;
-	else if (!strcmp(feat, "tun"))
-		chk_feature = check_tun;
-	else if (!strcmp(feat, "userns"))
-		chk_feature = check_userns;
-	else if (!strcmp(feat, "fdinfo_lock"))
-		chk_feature = check_fdinfo_lock;
-	else if (!strcmp(feat, "seccomp_suspend"))
-		chk_feature = check_ptrace_suspend_seccomp;
-	else if (!strcmp(feat, "seccomp_filters"))
-		chk_feature = check_ptrace_dump_seccomp_filters;
-	else if (!strcmp(feat, "loginuid"))
-		chk_feature = check_loginuid;
-	else if (!strcmp(feat, "cgroupns"))
-		chk_feature = check_cgroupns;
-	else {
-		pr_err("Unknown feature %s\n", feat);
-		return -1;
-	}
+	struct feature_list *fl;
 
-	return 0;
+	for (fl = feature_list; fl->name; fl++) {
+		if (!strcmp(feat, fl->name)) {
+			chk_feature = fl->func;
+			return 0;
+		}
+	}
+	pr_err("Unknown feature %s\n", feat);
+	return -1;
+}
+
+static char *feature_name(int (*func)())
+{
+	struct feature_list *fl;
+
+	for (fl = feature_list; fl->func; fl++) {
+		if (fl->func == func)
+			return fl->name;
+	}
+	return NULL;
 }
