@@ -28,6 +28,7 @@
 #include "net.h"
 #include "mount.h"
 #include "cgroup.h"
+#include "cgroup-props.h"
 #include "action-scripts.h"
 #include "sockets.h"
 #include "irmap.h"
@@ -37,6 +38,7 @@
 #include "setproctitle.h"
 
 #include "cr-errno.h"
+#include "namespaces.h"
 
 unsigned int service_sk_ino = -1;
 
@@ -108,6 +110,13 @@ err:
 	return -1;
 }
 
+static void set_resp_err(CriuResp *resp)
+{
+	resp->cr_errno = get_cr_errno();
+	resp->has_cr_errno = resp->cr_errno ? true : false;
+	resp->cr_errmsg = log_first_err();
+}
+
 static void send_criu_err(int sk, char *msg)
 {
 	CriuResp resp = CRIU_RESP__INIT;
@@ -116,10 +125,7 @@ static void send_criu_err(int sk, char *msg)
 
 	resp.type = CRIU_REQ_TYPE__EMPTY;
 	resp.success = false;
-	if (get_cr_errno()) {
-		resp.has_cr_errno = true;
-		resp.cr_errno = get_cr_errno();
-	}
+	set_resp_err(&resp);
 
 	send_criu_msg(sk, &resp);
 }
@@ -131,10 +137,7 @@ int send_criu_dump_resp(int socket_fd, bool success, bool restored)
 
 	msg.type = CRIU_REQ_TYPE__DUMP;
 	msg.success = success;
-	if (get_cr_errno()) {
-		msg.has_cr_errno = true;
-		msg.cr_errno = get_cr_errno();
-	}
+	set_resp_err(&msg);
 	msg.dump = &resp;
 
 	resp.has_restored = true;
@@ -149,10 +152,7 @@ static int send_criu_pre_dump_resp(int socket_fd, bool success)
 
 	msg.type = CRIU_REQ_TYPE__PRE_DUMP;
 	msg.success = success;
-	if (get_cr_errno()) {
-		msg.has_cr_errno = true;
-		msg.cr_errno = get_cr_errno();
-	}
+	set_resp_err(&msg);
 
 	return send_criu_msg(socket_fd, &msg);
 }
@@ -164,10 +164,7 @@ int send_criu_restore_resp(int socket_fd, bool success, int pid)
 
 	msg.type = CRIU_REQ_TYPE__RESTORE;
 	msg.success = success;
-	if (get_cr_errno()) {
-		msg.has_cr_errno = true;
-		msg.cr_errno = get_cr_errno();
-	}
+	set_resp_err(&msg);
 	msg.restore = &resp;
 
 	resp.pid = pid;
@@ -288,6 +285,11 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 		goto err;
 	}
 
+	if (log_keep_err()) {
+		pr_perror("Can't tune log");
+		goto err;
+	}
+
 	/* checking flags from client */
 	if (req->has_leave_running && req->leave_running)
 		opts.final_state = TASK_ALIVE;
@@ -320,6 +322,9 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 	if (req->has_tcp_established)
 		opts.tcp_established_ok = req->tcp_established;
 
+	if (req->has_tcp_skip_in_flight)
+		opts.tcp_skip_in_flight = req->tcp_skip_in_flight;
+
 	if (req->has_evasive_devices)
 		opts.evasive_devices = req->evasive_devices;
 
@@ -360,8 +365,7 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 		}
 	}
 
-	if (req->notify_scripts &&
-			add_script(SCRIPT_RPC_NOTIFY, sk))
+	if (req->notify_scripts && add_rpc_notify(sk))
 		goto err;
 
 	for (i = 0; i < req->n_veths; i++) {
@@ -374,6 +378,11 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 			goto err;
 	}
 
+	for (i = 0; i < req->n_join_ns; i++) {
+		if (join_ns_add(req->join_ns[i]->ns, req->join_ns[i]->ns_file, req->join_ns[i]->extra_opt))
+			goto err;
+	}
+
 	if (req->n_inherit_fd && !opts.swrk_restore) {
 		pr_err("inherit_fd is not allowed in standalone service\n");
 		goto err;
@@ -448,6 +457,23 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 		opts.manage_cgroups = mode;
 	}
 
+	if (req->freeze_cgroup)
+		opts.freeze_cgroup = req->freeze_cgroup;
+
+	if (req->has_timeout)
+		opts.timeout = req->timeout;
+
+	if (req->cgroup_props)
+		opts.cgroup_props = req->cgroup_props;
+
+	if (req->cgroup_props_file)
+		opts.cgroup_props_file = req->cgroup_props_file;
+
+	for (i = 0; i < req->n_cgroup_dump_controller; i++) {
+		if (!cgp_add_dump_controller(req->cgroup_dump_controller[i]))
+			goto err;
+	}
+
 	if (req->has_auto_ext_mnt)
 		opts.autodetect_ext_mounts = req->auto_ext_mnt;
 
@@ -473,6 +499,9 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 		}
 	}
 
+	if (check_namespace_opts())
+		goto err;
+
 	return 0;
 
 err:
@@ -567,9 +596,6 @@ static int check(int sk)
 
 	setproctitle("check --rpc");
 
-	/* Check only minimal kernel support */
-	opts.check_ms_kernel = true;
-
 	if (!cr_check())
 		resp.success = true;
 
@@ -752,7 +778,7 @@ static int chk_keepopen_req(CriuReq *msg)
 /*
  * Generic function to handle CRIU_REQ_TYPE__FEATURE_CHECK.
  *
- * The function will have resp.sucess = true for most cases
+ * The function will have resp.success = true for most cases
  * and the actual result will be in resp.features.
  *
  * For each feature which has been requested in msg->features
@@ -958,8 +984,8 @@ static void reap_worker(int signo)
 			pr_info("Worker(pid %d) exited with %d\n",
 				pid, WEXITSTATUS(status));
 		else if (WIFSIGNALED(status))
-			pr_info("Worker(pid %d) was killed by %d\n",
-				pid, WTERMSIG(status));
+			pr_info("Worker(pid %d) was killed by %d: %s\n", pid,
+				WTERMSIG(status), strsignal(WTERMSIG(status)));
 	}
 }
 
@@ -1074,7 +1100,7 @@ int cr_service(bool daemon_mode)
 
 		pr_info("Waiting for connection...\n");
 
-		sk = accept(server_fd, &client_addr, &client_addr_len);
+		sk = accept(server_fd, (struct sockaddr *)&client_addr, &client_addr_len);
 		if (sk == -1) {
 			pr_perror("Can't accept connection");
 			goto err;
