@@ -36,7 +36,9 @@
 #include "cr-service.h"
 #include "plugin.h"
 #include "mount.h"
+#include "namespaces.h"
 #include "cgroup.h"
+#include "cgroup-props.h"
 #include "cpu.h"
 #include "action-scripts.h"
 #include "irmap.h"
@@ -45,6 +47,7 @@
 #include "proc_parse.h"
 
 #include "setproctitle.h"
+#include "sysctl.h"
 
 struct cr_options opts;
 
@@ -56,10 +59,10 @@ void init_opts(void)
 	opts.final_state = TASK_DEAD;
 	INIT_LIST_HEAD(&opts.ext_unixsk_ids);
 	INIT_LIST_HEAD(&opts.veth_pairs);
-	INIT_LIST_HEAD(&opts.scripts);
 	INIT_LIST_HEAD(&opts.ext_mounts);
 	INIT_LIST_HEAD(&opts.inherit_fds);
 	INIT_LIST_HEAD(&opts.external);
+	INIT_LIST_HEAD(&opts.join_ns);
 	INIT_LIST_HEAD(&opts.new_cgroup_roots);
 	INIT_LIST_HEAD(&opts.irmap_scan_paths);
 
@@ -71,32 +74,27 @@ void init_opts(void)
 	opts.empty_ns = 0;
 }
 
-static int parse_ns_string(const char *ptr)
+static int parse_join_ns(const char *ptr)
 {
-	const char *end = ptr + strlen(ptr);
+	char *aux, *ns_file, *extra_opts = NULL;
+
+	aux = strchr(ptr, ':');
+	if (aux == NULL)
+		return -1;
+	*aux = '\0';
+
+	ns_file = aux + 1;
+	aux = strchr(ns_file, ',');
+	if (aux != NULL) {
+		*aux = '\0';
+		extra_opts = aux + 1;
+	} else {
+		extra_opts = NULL;
+	}
+	if (join_ns_add(ptr, ns_file, extra_opts))
+		return -1;
 
-	do {
-		if (ptr[3] != ',' && ptr[3] != '\0')
-			goto bad_ns;
-		if (!strncmp(ptr, "uts", 3))
-			opts.rst_namespaces_flags |= CLONE_NEWUTS;
-		else if (!strncmp(ptr, "ipc", 3))
-			opts.rst_namespaces_flags |= CLONE_NEWIPC;
-		else if (!strncmp(ptr, "mnt", 3))
-			opts.rst_namespaces_flags |= CLONE_NEWNS;
-		else if (!strncmp(ptr, "pid", 3))
-			opts.rst_namespaces_flags |= CLONE_NEWPID;
-		else if (!strncmp(ptr, "net", 3))
-			opts.rst_namespaces_flags |= CLONE_NEWNET;
-		else
-			goto bad_ns;
-		ptr += 4;
-	} while (ptr < end);
 	return 0;
-
-bad_ns:
-	pr_msg("Error: unknown namespace: %s\n", ptr);
-	return -1;
 }
 
 static int parse_cpu_cap(struct cr_options *opts, const char *optarg)
@@ -208,6 +206,17 @@ int add_external(char *key)
 	return 0;
 }
 
+bool deprecated_ok(char *what)
+{
+	if (opts.deprecated_ok)
+		return true;
+
+	pr_err("Deprecated functionality (%s) rejected.\n", what);
+	pr_err("Use the --deprecated option or set CRIU_DEPRECATED environment.\n");
+	pr_err("For details visit https://criu.org/Deprecation\n");
+	return false;
+}
+
 int main(int argc, char *argv[], char *envp[])
 {
 	pid_t pid = 0, tree_id = 0;
@@ -217,8 +226,7 @@ int main(int argc, char *argv[], char *envp[])
 	int opt, idx;
 	int log_level = LOG_UNSET;
 	char *imgs_dir = ".";
-	char *work_dir = NULL;
-	static const char short_opts[] = "dSsRf:F:t:p:hcD:o:n:v::x::Vr:jlW:L:M:";
+	static const char short_opts[] = "dSsRf:F:t:p:hcD:o:v::x::Vr:jJ:lW:L:M:";
 	static struct option long_opts[] = {
 		{ "tree",			required_argument,	0, 't'	},
 		{ "pid",			required_argument,	0, 'p'	},
@@ -233,7 +241,7 @@ int main(int argc, char *argv[], char *envp[])
 		{ "images-dir",			required_argument,	0, 'D'	},
 		{ "work-dir",			required_argument,	0, 'W'	},
 		{ "log-file",			required_argument,	0, 'o'	},
-		{ "namespaces",			required_argument,	0, 'n'	},
+		{ "join-ns",			required_argument,	0, 'J'	},
 		{ "root",			required_argument,	0, 'r'	},
 		{ USK_EXT_PARAM,		optional_argument,	0, 'x'	},
 		{ "help",			no_argument,		0, 'h'	},
@@ -275,10 +283,20 @@ int main(int argc, char *argv[], char *envp[])
 		{ "timeout",			required_argument,	0, 1072 },
 		{ "external",			required_argument,	0, 1073	},
 		{ "empty-ns",			required_argument,	0, 1074	},
+		{ "extra",			no_argument,		0, 1077	},
+		{ "experimental",		no_argument,		0, 1078	},
+		{ "all",			no_argument,		0, 1079	},
+		{ "cgroup-props",		required_argument,	0, 1080	},
+		{ "cgroup-props-file",		required_argument,	0, 1081	},
+		{ "cgroup-dump-controller",	required_argument,	0, 1082	},
+		{ SK_INFLIGHT_PARAM,		no_argument,		0, 1083	},
+		{ "deprecated",			no_argument,		0, 1084 },
 		{ },
 	};
 
 	BUILD_BUG_ON(PAGE_SIZE != PAGE_IMAGE_SIZE);
+	BUILD_BUG_ON(CTL_32 != SYSCTL_TYPE__CTL_32);
+	BUILD_BUG_ON(__CTL_STR != SYSCTL_TYPE__CTL_STR);
 
 	if (fault_injection_init())
 		return 1;
@@ -357,13 +375,13 @@ int main(int argc, char *argv[], char *envp[])
 			imgs_dir = optarg;
 			break;
 		case 'W':
-			work_dir = optarg;
+			opts.work_dir = optarg;
 			break;
 		case 'o':
 			opts.output = optarg;
 			break;
-		case 'n':
-			if (parse_ns_string(optarg))
+		case 'J':
+			if (parse_join_ns(optarg))
 				goto bad_arg;
 			break;
 		case 'v':
@@ -417,7 +435,7 @@ int main(int argc, char *argv[], char *envp[])
 			}
 			break;
 		case 1049:
-			if (add_script(optarg, 0))
+			if (add_script(optarg))
 				return 1;
 
 			break;
@@ -455,8 +473,8 @@ int main(int argc, char *argv[], char *envp[])
 			opts.force_irmap = true;
 			break;
 		case 1054:
-			opts.check_ms_kernel = true;
-			break;
+			pr_err("--ms is deprecated; see \"Check options\" of criu --help\n");
+			return 1;
 		case 'L':
 			opts.libdir = optarg;
 			break;
@@ -490,8 +508,11 @@ int main(int argc, char *argv[], char *envp[])
 				return 1;
 			break;
 		case 1063:
-			if (check_add_feature(optarg) < 0)
+			ret = check_add_feature(optarg);
+			if (ret < 0)	/* invalid kernel feature name */
 				return 1;
+			if (ret > 0)	/* list kernel features and exit */
+				return 0;
 			break;
 		case 1064:
 			if (!add_skip_mount(optarg))
@@ -554,6 +575,34 @@ int main(int argc, char *argv[], char *envp[])
 				return 1;
 			}
 			break;
+		case 1077:
+			opts.check_extra_features = true;
+			break;
+		case 1078:
+			opts.check_experimental_features = true;
+			break;
+		case 1079:
+			opts.check_extra_features = true;
+			opts.check_experimental_features = true;
+			break;
+		case 1080:
+			opts.cgroup_props = optarg;
+			break;
+		case 1081:
+			opts.cgroup_props_file = optarg;
+			break;
+		case 1082:
+			if (!cgp_add_dump_controller(optarg))
+				return 1;
+			break;
+		case 1083:
+			pr_msg("Will skip in-flight TCP connections\n");
+			opts.tcp_skip_in_flight = true;
+			break;
+		case 1084:
+			pr_msg("Turn deprecated stuff ON\n");
+			opts.deprecated_ok = true;
+			break;
 		case 'V':
 			pr_msg("Version: %s\n", CRIU_VERSION);
 			if (strcmp(CRIU_GITID, "0"))
@@ -567,6 +616,16 @@ int main(int argc, char *argv[], char *envp[])
 		}
 	}
 
+	if (getenv("CRIU_DEPRECATED")) {
+		pr_msg("Turn deprecated stuff ON via env\n");
+		opts.deprecated_ok = true;
+	}
+
+	if (check_namespace_opts()) {
+		pr_msg("Error: namespace flags confict\n");
+		return 1;
+	}
+
 	if (!opts.restore_detach && opts.restore_sibling) {
 		pr_msg("--restore-sibling only makes sense with --restore-detach\n");
 		return 1;
@@ -577,8 +636,8 @@ int main(int argc, char *argv[], char *envp[])
 		return 1;
 	}
 
-	if (work_dir == NULL)
-		work_dir = imgs_dir;
+	if (opts.work_dir == NULL)
+		opts.work_dir = imgs_dir;
 
 	if (optind >= argc) {
 		pr_msg("Error: command is required\n");
@@ -615,8 +674,18 @@ int main(int argc, char *argv[], char *envp[])
 			return 1;
 	}
 
-	if (chdir(work_dir)) {
-		pr_perror("Can't change directory to %s", work_dir);
+	/*
+	 * When a process group becomes an orphan,
+	 * its processes are sent a SIGHUP signal
+	 */
+	if (!strcmp(argv[optind], "restore") &&
+			opts.restore_detach &&
+			opts.final_state == TASK_STOPPED &&
+			opts.shell_job)
+		pr_warn("Stopped and detached shell job will get SIGHUP from OS.");
+
+	if (chdir(opts.work_dir)) {
+		pr_perror("Can't change directory to %s", opts.work_dir);
 		return 1;
 	}
 
@@ -625,10 +694,9 @@ int main(int argc, char *argv[], char *envp[])
 	if (log_init(opts.output))
 		return 1;
 
-	if (!list_empty(&opts.external) && strcmp(argv[optind], "dump")) {
-		pr_err("--external is dump-only option\n");
-		return 1;
-	}
+	pr_debug("Version: %s (gitid %s)\n", CRIU_VERSION, CRIU_GITID);
+	if (opts.deprecated_ok)
+		pr_debug("DEPRECATED ON\n");
 
 	if (!list_empty(&opts.inherit_fds)) {
 		if (strcmp(argv[optind], "restore")) {
@@ -715,7 +783,7 @@ usage:
 "Usage:\n"
 "  criu dump|pre-dump -t PID [<options>]\n"
 "  criu restore [<options>]\n"
-"  criu check [--ms]\n"
+"  criu check [--feature FEAT]\n"
 "  criu exec -p PID <syscall-string>\n"
 "  criu page-server\n"
 "  criu service [<options>]\n"
@@ -752,16 +820,19 @@ usage:
 "     --pidfile FILE     write root task, service or page-server pid to FILE\n"
 "  -W|--work-dir DIR     directory to cd and write logs/pidfiles/stats to\n"
 "                        (if not specified, value of --images-dir is used)\n"
-"     --cpu-cap [CAP]    require certain cpu capability. CAP: may be one of:\n"
-"                        'cpu','fpu','all','ins','none'. To disable capability, prefix it with '^'.\n"
+"     --cpu-cap [CAP]    CPU capabilities to write/check. CAP is comma-separated\n"
+"                        list of: cpu, fpu, all, ins, none. To disable\n"
+"                        a capability, use ^CAP. Empty argument implies all\n"
 "     --exec-cmd         execute the command specified after '--' on successful\n"
 "                        restore making it the parent of the restored process\n"
-"  --freeze-cgroup\n"
-"                        use cgroup freezer to collect processes\n"
+"  --freeze-cgroup       use cgroup freezer to collect processes\n"
 "\n"
 "* Special resources support:\n"
-"  -x|--" USK_EXT_PARAM "inode,.." "      allow external unix connections (optionally can be assign socket's inode that allows one-sided dump)\n"
+"  -x|--" USK_EXT_PARAM " [inode,...]\n"
+"                        allow external unix connections (optional arguments\n"
+"                        are socketpair inode(s) that allow one-sided dump)\n"
 "     --" SK_EST_PARAM "  checkpoint/restore established TCP connections\n"
+"     --" SK_INFLIGHT_PARAM "   skip (ignore) in-flight TCP connections\n"
 "  -r|--root PATH        change the root filesystem (when run in mount namespace)\n"
 "  --evasive-devices     use any path to a device file if the original one\n"
 "                        is inaccessible\n"
@@ -769,7 +840,7 @@ usage:
 "                        can optionally append @<bridge-name> to OUT for moving\n"
 "                        the outside veth to the named bridge\n"
 "  --link-remap          allow one to link unlinked files back when possible\n"
-"  --ghost-limit size    specify maximum size of deleted file contents to be carried inside an image file\n"
+"  --ghost-limit size    limit max size of deleted file contents inside image\n"
 "  --action-script FILE  add an external action script\n"
 "  -j|--" OPT_SHELL_JOB "        allow one to dump and restore shell jobs\n"
 "  -l|--" OPT_FILE_LOCKS "       handle file locks, for safety, only used for container\n"
@@ -785,31 +856,58 @@ usage:
 "                        allow autoresolving mounts with external sharing\n"
 "  --enable-external-masters\n"
 "                        allow autoresolving mounts with external masters\n"
-"  --manage-cgroups [m]  dump or restore cgroups the process is in usig mode:\n"
-"                        'none', 'props', 'soft' (default), 'full' and 'strict'.\n"
+"  --manage-cgroups [m]  dump/restore process' cgroups; argument can be one of\n"
+"                        'none', 'props', 'soft' (default), 'full' or 'strict'\n"
 "  --cgroup-root [controller:]/newroot\n"
 "                        change the root cgroup the controller will be\n"
 "                        installed into. No controller means that root is the\n"
-"                        default for all controllers not specified.\n"
-"  --skip-mnt PATH       ignore this mountpoint when dumping the mount namespace.\n"
-"  --enable-fs FSNAMES   a comma separated list of filesystem names or \"all\".\n"
+"                        default for all controllers not specified\n"
+"  --cgroup-props STRING\n"
+"                        define cgroup controllers and properties\n"
+"                        to be checkpointed, which are described\n"
+"                        via STRING using simplified YAML format\n"
+"  --cgroup-props-file FILE\n"
+"                        same as --cgroup-props, but taking description\n"
+"                        from the path specified\n"
+"  --cgroup-dump-controller NAME\n"
+"                        define cgroup controller to be dumped\n"
+"                        and skip anything else present in system\n"
+"  --skip-mnt PATH       ignore this mountpoint when dumping the mount namespace\n"
+"  --enable-fs FSNAMES   a comma separated list of filesystem names or \"all\"\n"
 "                        force criu to (try to) dump/restore these filesystem's\n"
-"                        mountpoints even if fs is not supported.\n"
+"                        mountpoints even if fs is not supported\n"
 "  --external RES        dump objects from this list as external resources:\n"
-"                        Formats of RES:\n"
+"                        Formats of RES on dump:\n"
 "                            tty[rdev:dev]\n"
 "                            file[mnt_id:inode]\n"
-"  --inherit-fd fd[<num>]:<existing>\n"
-"                        Inherit file descriptors. This allows to treat file descriptor\n"
-"                        <num> as being already opened via <existing> one and instead of\n"
-"                        trying to open we inherit it:\n"
+"                            dev[maj:min]:VAL\n"
+"                        Formats of RES on restore:\n"
+"                            dev[VAL]:DEVPATH\n"
+"  --inherit-fd fd[NUM]:RES\n"
+"                        Inherit file descriptors, treating fd NUM as being\n"
+"                        already opened via an existing RES, which can be:\n"
 "                            tty[rdev:dev]\n"
 "                            pipe[inode]\n"
 "                            socket[inode]\n"
 "                            file[mnt_id:inode]\n"
-"  --empty-ns {net}\n"
-"			Create a namespace, but don't restore its properies.\n"
-"			An user will retore them from action scripts.\n"
+"  --empty-ns net        Create a namespace, but don't restore its properies\n"
+"                        (assuming it will be restored by action scripts)\n"
+"  -J|--join-ns NS:{PID|NS_FILE}[,OPTIONS]\n"
+"			Join existing namespace and restore process in it.\n"
+"			Namespace can be specified as either pid or file path.\n"
+"			OPTIONS can be used to specify parameters for userns:\n"
+"			    user:PID,UID,GID\n"
+"\n"
+"Check options:\n"
+"  Without options, \"criu check\" checks availability of absolutely required\n"
+"  kernel features, critical for performing dump and restore.\n"
+"  --extra               add check for extra kernel features\n"
+"  --experimental        add check for experimental kernel features\n"
+"  --all                 same as --extra --experimental\n"
+"  --feature FEAT        only check a particular feature, one of:"
+	);
+	pr_check_features("                            ", ", ", 80);
+	pr_msg(
 "\n"
 "* Logging:\n"
 "  -o|--log-file FILE    log file name\n"
@@ -827,7 +925,7 @@ usage:
 "  --auto-dedup          when used on dump it will deduplicate \"old\" data in\n"
 "                        pages images of previous dump\n"
 "                        when used on restore, as soon as page is restored, it\n"
-"                        will be punched from the image.\n"
+"                        will be punched from the image\n"
 "\n"
 "Page/Service server options:\n"
 "  --address ADDR        address of server or service\n"
@@ -837,7 +935,6 @@ usage:
 "Other options:\n"
 "  -h|--help             show this text\n"
 "  -V|--version          show version\n"
-"     --ms               don't check not yet merged kernel features\n"
 	);
 
 	return 0;
