@@ -16,7 +16,7 @@
 #include "asm/types.h"
 #include "imgset.h"
 #include "eventfd.h"
-#include "proc_parse.h"
+#include "fdinfo.h"
 #include "image.h"
 #include "util.h"
 #include "log.h"
