@@ -19,7 +19,7 @@
 #include "file-lock.h"
 #include "image.h"
 #include "list.h"
-#include "util.h"
+#include "rst-malloc.h"
 #include "util-pie.h"
 #include "lock.h"
 #include "sockets.h"
@@ -37,8 +37,9 @@
 #include "timerfd.h"
 #include "imgset.h"
 #include "fs-magic.h"
-#include "proc_parse.h"
+#include "fdinfo.h"
 #include "cr_options.h"
+#include "autofs.h"
 
 #include "parasite.h"
 #include "parasite-syscall.h"
@@ -96,6 +97,45 @@ static inline struct file_desc *find_file_desc(FdinfoEntry *fe)
 	return find_file_desc_raw(fe->type, fe->id);
 }
 
+struct fdinfo_list_entry *find_used_fd(struct list_head *head, int fd)
+{
+	struct fdinfo_list_entry *fle;
+
+	list_for_each_entry_reverse(fle, head, used_list) {
+		if (fle->fe->fd == fd)
+			return fle;
+		/* List is ordered, so let's stop */
+		if (fle->fe->fd < fd)
+			break;
+	}
+	return NULL;
+}
+
+unsigned int find_unused_fd(struct list_head *head, int hint_fd)
+{
+	struct fdinfo_list_entry *fle;
+	int fd = 0, prev_fd;
+
+	if ((hint_fd >= 0) && (!find_used_fd(head, hint_fd))) {
+		fd = hint_fd;
+		goto out;
+	}
+
+	prev_fd = service_fd_min_fd() - 1;
+
+	list_for_each_entry_reverse(fle, head, used_list) {
+		fd = fle->fe->fd;
+		if (prev_fd > fd) {
+			fd++;
+			goto out;
+		}
+		prev_fd = fd - 1;
+	}
+	BUG();
+out:
+	return fd;
+}
+
 /*
  * A file may be shared between several file descriptors. E.g
  * when doing a fork() every fd of a forker and respective fds
@@ -113,7 +153,7 @@ static inline struct file_desc *find_file_desc(FdinfoEntry *fe)
  * 2. Open step.
  *    The one who creates the file (the 'master') creates one,
  *    then creates one more unix socket (transport) and sends the
- *    created file over this socket to the other recepients.
+ *    created file over this socket to the other recipients.
  * 3. Receive step.
  *    Those, who wait for the file to appear, receive one via
  *    the transport socket, then close the socket and dup() the
@@ -238,7 +278,7 @@ int do_dump_gen_file(struct fd_parms *p, int lfd,
 	if (ret < 0)
 		return ret;
 
-	pr_info("fdinfo: type: 0x%2x flags: %#o/%#o pos: 0x%8"PRIx64" fd: %d\n",
+	pr_info("fdinfo: type: %#2x flags: %#o/%#o pos: %#8"PRIx64" fd: %d\n",
 		ops->type, p->flags, (int)p->fd_flags, p->pos, p->fd);
 
 	return pb_write_one(img, &e, PB_FDINFO);
@@ -295,7 +335,7 @@ static int fill_fd_params(struct parasite_ctl *ctl, int fd, int lfd,
 
 	fown_entry__init(&p->fown);
 
-	pr_info("%d fdinfo %d: pos: 0x%16"PRIx64" flags: %16o/%#x\n",
+	pr_info("%d fdinfo %d: pos: %#16"PRIx64" flags: %16o/%#x\n",
 		ctl->pid.real, fd, p->pos, p->flags, (int)p->fd_flags);
 
 	ret = fcntl(lfd, F_GETSIG, 0);
@@ -321,6 +361,8 @@ static const struct fdtype_ops *get_misc_dev_ops(int minor)
 	switch (minor) {
 	case TUN_MINOR:
 		return &tunfile_dump_ops;
+	case AUTOFS_MINOR:
+		return &regfile_dump_ops;
 	};
 
 	return NULL;
@@ -389,9 +431,10 @@ static int dump_one_file(struct parasite_ctl *ctl, int fd, int lfd, struct fd_op
 {
 	struct fd_parms p = FD_PARMS_INIT;
 	const struct fdtype_ops *ops;
+	struct fd_link link;
 
 	if (fill_fd_params(ctl, fd, lfd, opts, &p) < 0) {
-		pr_perror("Can't get stat on %d", fd);
+		pr_err("Can't get stat on %d\n", fd);
 		return -1;
 	}
 
@@ -429,8 +472,6 @@ static int dump_one_file(struct parasite_ctl *ctl, int fd, int lfd, struct fd_op
 	}
 
 	if (S_ISREG(p.stat.st_mode) || S_ISDIR(p.stat.st_mode)) {
-		struct fd_link link;
-
 		if (fill_fdlink(lfd, &p, &link))
 			return -1;
 
@@ -453,52 +494,67 @@ static int dump_one_file(struct parasite_ctl *ctl, int fd, int lfd, struct fd_op
 		return do_dump_gen_file(&p, lfd, ops, img);
 	}
 
-	return dump_unsupp_fd(&p, lfd, img, "unknown", NULL);
+	/*
+	 * For debug purpose -- at least show the link
+	 * file pointing to when reporting unsupported file.
+	 * On error simply empty string here.
+	 */
+	if (fill_fdlink(lfd, &p, &link))
+		memzero(&link, sizeof(link));
+
+	return dump_unsupp_fd(&p, lfd, img, "unknown", link.name + 1);
 }
 
 int dump_task_files_seized(struct parasite_ctl *ctl, struct pstree_item *item,
 		struct parasite_drain_fd *dfds)
 {
-	int *lfds;
-	struct cr_img *img;
-	struct fd_opts *opts;
+	int *lfds = NULL;
+	struct cr_img *img = NULL;
+	struct fd_opts *opts = NULL;
 	int i, ret = -1;
+	int off, nr_fds = min((int) PARASITE_MAX_FDS, dfds->nr_fds);
 
 	pr_info("\n");
 	pr_info("Dumping opened files (pid: %d)\n", ctl->pid.real);
 	pr_info("----------------------------------------\n");
 
-	lfds = xmalloc(dfds->nr_fds * sizeof(int));
+	lfds = xmalloc(nr_fds * sizeof(int));
 	if (!lfds)
 		goto err;
 
-	opts = xmalloc(dfds->nr_fds * sizeof(struct fd_opts));
+	opts = xmalloc(nr_fds * sizeof(struct fd_opts));
 	if (!opts)
-		goto err1;
-
-	ret = parasite_drain_fds_seized(ctl, dfds, lfds, opts);
-	if (ret)
-		goto err2;
+		goto err;
 
 	img = open_image(CR_FD_FDINFO, O_DUMP, item->ids->files_id);
 	if (!img)
-		goto err2;
+		goto err;
 
-	for (i = 0; i < dfds->nr_fds; i++) {
-		ret = dump_one_file(ctl, dfds->fds[i], lfds[i], opts + i, img);
-		close(lfds[i]);
+	ret = 0; /* Don't fail if nr_fds == 0 */
+	for (off = 0; off < dfds->nr_fds; off += nr_fds) {
+		if (nr_fds + off > dfds->nr_fds)
+			nr_fds = dfds->nr_fds - off;
+
+		ret = parasite_drain_fds_seized(ctl, dfds, nr_fds,
+							off, lfds, opts);
 		if (ret)
-			break;
+			goto err;
+
+		for (i = 0; i < nr_fds; i++) {
+			ret = dump_one_file(ctl, dfds->fds[i + off],
+						lfds[i], opts + i, img);
+			close(lfds[i]);
+			if (ret)
+				break;
+		}
 	}
 
-	close_image(img);
-
 	pr_info("----------------------------------------\n");
-err2:
-	xfree(opts);
-err1:
-	xfree(lfds);
 err:
+	if (img)
+		close_image(img);
+	xfree(opts);
+	xfree(lfds);
 	return ret;
 }
 
@@ -643,12 +699,43 @@ static int collect_fd(int pid, FdinfoEntry *e, struct rst_info *rst_info)
 	else
 		collect_gen_fd(new_le, rst_info);
 
+	collect_used_fd(new_le, rst_info);
+
 	list_add_tail(&new_le->desc_list, &le->desc_list);
 	new_le->desc = fdesc;
 
 	return 0;
 }
 
+FdinfoEntry *dup_fdinfo(FdinfoEntry *old, int fd, unsigned flags)
+{
+	FdinfoEntry *e;
+
+	e = shmalloc(sizeof(*e));
+	if (!e)
+		return NULL;
+
+	fdinfo_entry__init(e);
+
+	e->id		= old->id;
+	e->type		= old->type;
+	e->fd		= fd;
+	e->flags	= flags;
+	return e;
+}
+
+int dup_fle(struct pstree_item *task, struct fdinfo_list_entry *ple,
+		   int fd, unsigned flags)
+{
+	FdinfoEntry *e;
+
+	e = dup_fdinfo(ple->fe, fd, flags);
+	if (!e)
+		return -1;
+
+	return collect_fd(task->pid.virt, e, rsti(task));
+}
+
 int prepare_ctl_tty(int pid, struct rst_info *rst_info, u32 ctl_tty_id)
 {
 	FdinfoEntry *e;
@@ -683,26 +770,21 @@ int prepare_fd_pid(struct pstree_item *item)
 	pid_t pid = item->pid.virt;
 	struct rst_info *rst_info = rsti(item);
 
+	INIT_LIST_HEAD(&rst_info->used);
 	INIT_LIST_HEAD(&rst_info->fds);
 	INIT_LIST_HEAD(&rst_info->eventpoll);
 	INIT_LIST_HEAD(&rst_info->tty_slaves);
 	INIT_LIST_HEAD(&rst_info->tty_ctty);
 
-	if (!fdinfo_per_id) {
-		img = open_image(CR_FD_FDINFO, O_RSTR, pid);
-		if (!img)
-			return -1;
-	} else {
-		if (item->ids == NULL) /* zombie */
-			return 0;
+	if (item->ids == NULL) /* zombie */
+		return 0;
 
-		if (rsti(item)->fdt && rsti(item)->fdt->pid != item->pid.virt)
-			return 0;
+	if (rsti(item)->fdt && rsti(item)->fdt->pid != item->pid.virt)
+		return 0;
 
-		img = open_image(CR_FD_FDINFO, O_RSTR, item->ids->files_id);
-		if (!img)
-			return -1;
-	}
+	img = open_image(CR_FD_FDINFO, O_RSTR, item->ids->files_id);
+	if (!img)
+		return -1;
 
 	while (1) {
 		FdinfoEntry *e;
@@ -711,6 +793,12 @@ int prepare_fd_pid(struct pstree_item *item)
 		if (ret <= 0)
 			break;
 
+		if (e->fd >= service_fd_min_fd()) {
+			ret = -1;
+			pr_err("Too big FD number to restore %d\n", e->fd);
+			break;
+		}
+
 		ret = collect_fd(pid, e, rst_info);
 		if (ret < 0) {
 			fdinfo_entry__free_unpacked(e, NULL);
@@ -835,7 +923,7 @@ static int open_transport_fd(int pid, struct fdinfo_list_entry *fle)
 		pr_perror("Can't create socket");
 		return -1;
 	}
-	ret = bind(sock, &saddr, sun_len);
+	ret = bind(sock, (struct sockaddr *)&saddr, sun_len);
 	if (ret < 0) {
 		pr_perror("Can't bind unix socket %s", saddr.sun_path + 1);
 		goto err;
@@ -868,7 +956,7 @@ int send_fd_to_peer(int fd, struct fdinfo_list_entry *fle, int sock)
 	return send_fd(sock, &saddr, len, fd);
 }
 
-static int send_fd_to_self(int fd, struct fdinfo_list_entry *fle, int *sock)
+static int send_fd_to_self(int fd, struct fdinfo_list_entry *fle, int sock)
 {
 	int dfd = fle->fe->fd;
 
@@ -879,10 +967,9 @@ static int send_fd_to_self(int fd, struct fdinfo_list_entry *fle, int *sock)
 	if (inherit_fd_resolve_clash(dfd) < 0)
 		return -1;
 
-	pr_info("\t\t\tGoing to dup %d into %d\n", fd, dfd);
-	if (move_img_fd(sock, dfd))
-		return -1;
+	BUG_ON(dfd == sock);
 
+	pr_info("\t\t\tGoing to dup %d into %d\n", fd, dfd);
 	if (dup2(fd, dfd) != dfd) {
 		pr_perror("Can't dup local fd %d -> %d", fd, dfd);
 		return -1;
@@ -918,17 +1005,13 @@ static int serve_out_fd(int pid, int fd, struct file_desc *d)
 	int sock, ret;
 	struct fdinfo_list_entry *fle;
 
-	sock = socket(PF_UNIX, SOCK_DGRAM, 0);
-	if (sock < 0) {
-		pr_perror("Can't create socket");
-		return -1;
-	}
+	sock = get_service_fd(TRANSPORT_FD_OFF);
 
 	pr_info("\t\tCreate fd for %d\n", fd);
 
 	list_for_each_entry(fle, &d->fd_info_head, desc_list) {
 		if (pid == fle->pid)
-			ret = send_fd_to_self(fd, fle, &sock);
+			ret = send_fd_to_self(fd, fle, sock);
 		else
 			ret = send_fd_to_peer(fd, fle, sock);
 
@@ -940,7 +1023,6 @@ static int serve_out_fd(int pid, int fd, struct file_desc *d)
 
 	ret = 0;
 out:
-	close(sock);
 	return ret;
 }
 
@@ -1143,26 +1225,18 @@ out:
 
 static int fchroot(int fd)
 {
-	char fd_path[PSFDS];
-	int proc;
-
 	/*
 	 * There's no such thing in syscalls. We can emulate
-	 * it using the /proc/self/fd/ :)
-	 *
-	 * But since there might be no /proc mount in our mount
-	 * namespace, we will have to ... workaround it.
+	 * it using fchdir()
 	 */
 
-	proc = get_service_fd(PROC_FD_OFF);
-	if (fchdir(proc) < 0) {
+	if (fchdir(fd) < 0) {
 		pr_perror("Can't chdir to proc");
 		return -1;
 	}
 
-	sprintf(fd_path, "./self/fd/%d", fd);
-	pr_debug("Going to chroot into %s\n", fd_path);
-	return chroot(fd_path);
+	pr_debug("Going to chroot into /proc/self/fd/%d\n", fd);
+	return chroot(".");
 }
 
 int restore_fs(struct pstree_item *me)
@@ -1188,9 +1262,8 @@ int restore_fs(struct pstree_item *me)
 	}
 
 	/*
-	 * Now do chroot/chdir. Chroot goes first as it
-	 * calls chdir into proc service descriptor so
-	 * we'd need to fix chdir after it anyway.
+	 * Now do chroot/chdir. Chroot goes first as it calls chdir into
+	 * dd_root so we'd need to fix chdir after it anyway.
 	 */
 
 	ret = fchroot(dd_root);
@@ -1585,3 +1658,35 @@ bool external_lookup_id(char *id)
 			return true;
 	return false;
 }
+
+char *external_lookup_by_key(char *key)
+{
+	struct external *ext;
+	int len = strlen(key);
+
+	list_for_each_entry(ext, &opts.external, node) {
+		if (strncmp(ext->id, key, len))
+			continue;
+		if (ext->id[len] == ':')
+			return ext->id + len + 1;
+	}
+	return NULL;
+}
+
+int open_transport_socket()
+{
+	int sock;
+
+	sock = socket(PF_UNIX, SOCK_DGRAM | SOCK_CLOEXEC, 0);
+	if (sock < 0) {
+		pr_perror("Can't create socket");
+		return -1;
+	}
+	if (install_service_fd(TRANSPORT_FD_OFF, sock) < 0) {
+		close(sock);
+		return -1;
+	}
+	close(sock);
+
+	return 0;
+}
