@@ -15,7 +15,7 @@
 #include <sys/vfs.h>
 #include <linux/magic.h>
 #include <sys/wait.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <sys/mman.h>
 #include <sys/mount.h>
 #include <aio.h>
@@ -26,7 +26,7 @@
 #include "asm/types.h"
 #include "imgset.h"
 #include "fsnotify.h"
-#include "proc_parse.h"
+#include "fdinfo.h"
 #include "mount.h"
 #include "image.h"
 #include "util.h"
@@ -40,6 +40,8 @@
 #include "cr_options.h"
 #include "namespaces.h"
 #include "pstree.h"
+#include "fault-injection.h"
+#include "syscall-codes.h"
 
 #include "protobuf.h"
 #include "images/fsnotify.pb-c.h"
@@ -103,7 +105,7 @@ static void decode_handle(fh_t *handle, FhEntry *img)
 
 static int open_by_handle(void *arg, int fd, int pid)
 {
-	return open_by_handle_at(fd, arg, O_PATH);
+	return syscall(__NR_open_by_handle_at, fd, arg, O_PATH);
 }
 
 static char *alloc_openable(unsigned int s_dev, unsigned long i_ino, FhEntry *f_handle)
@@ -162,7 +164,7 @@ static char *alloc_openable(unsigned int s_dev, unsigned long i_ino, FhEntry *f_
 		openable_fd = openat(mntfd, __path, O_PATH);
 		if (openable_fd >= 0) {
 			if (fstat(openable_fd, &st)) {
-				pr_perror("Can't stat on %s\n", __path);
+				pr_perror("Can't stat on %s", __path);
 				close(openable_fd);
 				return ERR_PTR(-errno);
 			}
@@ -203,14 +205,13 @@ static int open_handle(unsigned int s_dev, unsigned long i_ino,
 
 	mntfd = open_mount(s_dev);
 	if (mntfd < 0) {
-		pr_err("Mount root for 0x%08x not found\n", s_dev);
+		pr_err("Mount root for %#08x not found\n", s_dev);
 		goto out;
 	}
 
 	fd = userns_call(open_by_handle, UNS_FDOUT, &handle, sizeof(handle), mntfd);
 	if (fd < 0) {
-		errno = -fd;
-		pr_perror("Can't open file handle for 0x%08x:0x%016lx",
+		pr_perror("Can't open file handle for %#08x:%#016lx",
 				s_dev, i_ino);
 	}
 
@@ -225,7 +226,13 @@ int check_open_handle(unsigned int s_dev, unsigned long i_ino,
 	int fd = -1;
 	char *path;
 
+	if (fault_injected(FI_CHECK_OPEN_HANDLE)) {
+		fd = -1;
+		goto fault;
+	}
+
 	fd = open_handle(s_dev, i_ino, f_handle);
+fault:
 	if (fd >= 0) {
 		struct mount_info *mi;
 
@@ -301,9 +308,9 @@ static int dump_inotify_entry(union fdinfo_entries *e, void *arg)
 	struct inotify_wd_entry *wd_entry = (struct inotify_wd_entry *) e;
 	InotifyWdEntry *we = &wd_entry->e;
 
-	pr_info("wd: wd 0x%08x s_dev 0x%08x i_ino 0x%16"PRIx64" mask 0x%08x\n",
+	pr_info("wd: wd %#08x s_dev %#08x i_ino %#16"PRIx64" mask %#08x\n",
 			we->wd, we->s_dev, we->i_ino, we->mask);
-	pr_info("\t[fhandle] bytes 0x%08x type 0x%08x __handle 0x%016"PRIx64":0x%016"PRIx64"\n",
+	pr_info("\t[fhandle] bytes %#08x type %#08x __handle %#016"PRIx64":%#016"PRIx64"\n",
 			we->f_handle->bytes, we->f_handle->type,
 			we->f_handle->handle[0], we->f_handle->handle[1]);
 
@@ -333,7 +340,7 @@ static int dump_one_inotify(int lfd, u32 id, const struct fd_parms *p)
 	if (ret < 0)
 		return -1;
 	else if (ret > 0)
-		pr_warn("The 0x%08x inotify events will be dropped\n", id);
+		pr_warn("The %#08x inotify events will be dropped\n", id);
 
 	ie.id = id;
 	ie.flags = p->flags;
@@ -351,7 +358,7 @@ static int dump_one_inotify(int lfd, u32 id, const struct fd_parms *p)
 		ie.wd[i++] = &we->ify.e;
 	ie.n_wd = wd_list.n;
 
-	pr_info("id 0x%08x flags 0x%08x\n", ie.id, ie.flags);
+	pr_info("id %#08x flags %#08x\n", ie.id, ie.flags);
 	if (pb_write_one(img_from_set(glob_imgset, CR_FD_INOTIFY_FILE), &ie, PB_INOTIFY_FILE))
 		goto free;
 
@@ -395,10 +402,10 @@ static int dump_fanotify_entry(union fdinfo_entries *e, void *arg)
 
 		BUG_ON(!fme->ie);
 
-		pr_info("mark: s_dev 0x%08x i_ino 0x%016"PRIx64" mask 0x%08x\n",
+		pr_info("mark: s_dev %#08x i_ino %#016"PRIx64" mask %#08x\n",
 			fme->s_dev, fme->ie->i_ino, fme->mask);
 
-		pr_info("\t[fhandle] bytes 0x%08x type 0x%08x __handle 0x%016"PRIx64":0x%016"PRIx64"\n",
+		pr_info("\t[fhandle] bytes %#08x type %#08x __handle %#016"PRIx64":%#016"PRIx64"\n",
 			fme->ie->f_handle->bytes, fme->ie->f_handle->type,
 			fme->ie->f_handle->handle[0], fme->ie->f_handle->handle[1]);
 
@@ -418,7 +425,7 @@ static int dump_fanotify_entry(union fdinfo_entries *e, void *arg)
 		}
 		fme->s_dev = m->s_dev;
 
-		pr_info("mark: s_dev 0x%08x mnt_id  0x%08x mask 0x%08x\n",
+		pr_info("mark: s_dev %#08x mnt_id  %#08x mask %#08x\n",
 			fme->s_dev, fme->me->mnt_id, fme->mask);
 
 	}
@@ -443,7 +450,7 @@ static int dump_one_fanotify(int lfd, u32 id, const struct fd_parms *p)
 	if (ret < 0)
 		return -1;
 	else if (ret > 0)
-		pr_warn("The 0x%08x fanotify events will be dropped\n", id);
+		pr_warn("The %#08x fanotify events will be dropped\n", id);
 	ret = -1;
 
 	fe.id = id;
@@ -463,7 +470,7 @@ static int dump_one_fanotify(int lfd, u32 id, const struct fd_parms *p)
 		fe.mark[i++] = &we->ffy.e;
 	fe.n_mark = wd_list.n;
 
-	pr_info("id 0x%08x flags 0x%08x\n", fe.id, fe.flags);
+	pr_info("id %#08x flags %#08x\n", fe.id, fe.flags);
 
 	fe.faflags = wd_list.fsn_params.faflags;
 	fe.evflags = wd_list.fsn_params.evflags;
@@ -512,7 +519,7 @@ static char *get_mark_path(const char *who, struct file_remap *remap,
 
 		mntns_root = mntns_get_root_by_mnt_id(remap->rmnt_id);
 
-		pr_debug("\t\tRestore %s watch for 0x%08x:0x%016lx (via %s)\n",
+		pr_debug("\t\tRestore %s watch for %#08x:%#016lx (via %s)\n",
 			 who, s_dev, i_ino, remap->rpath);
 		*target = openat(mntns_root, remap->rpath, O_PATH);
 	} else if (f_handle->path) {
@@ -555,7 +562,7 @@ static char *get_mark_path(const char *who, struct file_remap *remap,
 		if (read_fd_link(*target, link, sizeof(link)) < 0)
 			link[0] = '\0';
 
-		pr_debug("\t\tRestore %s watch for 0x%08x:0x%016lx (via %s -> %s)\n",
+		pr_debug("\t\tRestore %s watch for %#08x:%#016lx (via %s -> %s)\n",
 				who, s_dev, i_ino, path, link);
 	}
 err:
@@ -683,12 +690,12 @@ static int open_inotify_fd(struct file_desc *d)
 
 	tmp = inotify_init1(info->ife->flags);
 	if (tmp < 0) {
-		pr_perror("Can't create inotify for 0x%08x", info->ife->id);
+		pr_perror("Can't create inotify for %#08x", info->ife->id);
 		return -1;
 	}
 
 	list_for_each_entry(wd_info, &info->marks, list) {
-		pr_info("\tRestore 0x%x wd for 0x%08x\n", wd_info->iwe->wd, wd_info->iwe->id);
+		pr_info("\tRestore 0x%x wd for %#08x\n", wd_info->iwe->wd, wd_info->iwe->id);
 		if (restore_one_inotify(tmp, wd_info)) {
 			close_safe(&tmp);
 			break;
@@ -718,13 +725,12 @@ static int open_fanotify_fd(struct file_desc *d)
 
 	ret = fanotify_init(flags, info->ffe->evflags);
 	if (ret < 0) {
-		errno = -ret;
 		pr_perror("Can't init fanotify mark (%d)", ret);
 		return -1;
 	}
 
 	list_for_each_entry(mark, &info->marks, list) {
-		pr_info("\tRestore fanotify for 0x%08x\n", mark->fme->id);
+		pr_info("\tRestore fanotify for %#08x\n", mark->fme->id);
 		if (restore_one_fanotify(ret, mark)) {
 			close_safe(&ret);
 			break;
@@ -758,7 +764,7 @@ static struct fsnotify_file_info *find_inotify_info(unsigned id)
 		 * wd-s for one inotify in one row, thus sometimes
 		 * we can avoid scanning the inotify_info_head.
 		 */
-		pr_debug("\t\tlast ify for 0x%08x found\n", id);
+		pr_debug("\t\tlast ify for %#08x found\n", id);
 		return last;
 	}
 
@@ -768,7 +774,7 @@ static struct fsnotify_file_info *find_inotify_info(unsigned id)
 			return p;
 		}
 
-	pr_err("Can't find inotify with id 0x%08x\n", id);
+	pr_err("Can't find inotify with id %#08x\n", id);
 	return NULL;
 }
 
@@ -819,11 +825,11 @@ static int collect_fanotify_mark(struct fsnotify_mark_info *mark)
 			return __collect_inotify_mark(p, mark);
 	}
 
-	pr_err("Can't find fanotify with id 0x%08x\n", mark->fme->id);
+	pr_err("Can't find fanotify with id %#08x\n", mark->fme->id);
 	return -1;
 }
 
-static int collect_one_inotify(void *o, ProtobufCMessage *msg)
+static int collect_one_inotify(void *o, ProtobufCMessage *msg, struct cr_img *img)
 {
 	struct fsnotify_file_info *info = o;
 	int i;
@@ -831,7 +837,7 @@ static int collect_one_inotify(void *o, ProtobufCMessage *msg)
 	info->ife = pb_msg(msg, InotifyFileEntry);
 	INIT_LIST_HEAD(&info->marks);
 	list_add(&info->list, &inotify_info_head);
-	pr_info("Collected id 0x%08x flags 0x%08x\n", info->ife->id, info->ife->flags);
+	pr_info("Collected id %#08x flags %#08x\n", info->ife->id, info->ife->flags);
 
 	for (i = 0; i < info->ife->n_wd; i++) {
 		struct fsnotify_mark_info *mark;
@@ -858,7 +864,7 @@ struct collect_image_info inotify_cinfo = {
 	.collect	= collect_one_inotify,
 };
 
-static int collect_one_fanotify(void *o, ProtobufCMessage *msg)
+static int collect_one_fanotify(void *o, ProtobufCMessage *msg, struct cr_img *img)
 {
 	struct fsnotify_file_info *info = o;
 	int i;
@@ -866,7 +872,7 @@ static int collect_one_fanotify(void *o, ProtobufCMessage *msg)
 	info->ffe = pb_msg(msg, FanotifyFileEntry);
 	INIT_LIST_HEAD(&info->marks);
 	list_add(&info->list, &fanotify_info_head);
-	pr_info("Collected id 0x%08x flags 0x%08x\n", info->ffe->id, info->ffe->flags);
+	pr_info("Collected id %#08x flags %#08x\n", info->ffe->id, info->ffe->flags);
 
 	for (i = 0; i < info->ffe->n_mark; i++) {
 		struct fsnotify_mark_info *mark;
@@ -893,7 +899,7 @@ struct collect_image_info fanotify_cinfo = {
 	.collect	= collect_one_fanotify,
 };
 
-static int collect_one_inotify_mark(void *o, ProtobufCMessage *msg)
+static int collect_one_inotify_mark(void *o, ProtobufCMessage *msg, struct cr_img *i)
 {
 	struct fsnotify_mark_info *mark = o;
 
@@ -921,7 +927,7 @@ struct collect_image_info inotify_mark_cinfo = {
 	.collect	= collect_one_inotify_mark,
 };
 
-static int collect_one_fanotify_mark(void *o, ProtobufCMessage *msg)
+static int collect_one_fanotify_mark(void *o, ProtobufCMessage *msg, struct cr_img *i)
 {
 	struct fsnotify_mark_info *mark = o;
 
