@@ -9,6 +9,8 @@ extern int log_init(const char *output);
 extern void log_fini(void);
 extern int log_init_by_pid(void);
 extern void log_closedir(void);
+extern int log_keep_err(void);
+extern char *log_first_err(void);
 
 extern void log_set_fd(int fd);
 extern int log_get_fd(void);
