@@ -19,6 +19,7 @@ enum sfd_type {
 	CGROUP_YARD,
 	USERNSD_SK,	/* Socket for usernsd */
 	NS_FD_OFF,	/* Node's net namespace fd */
+	TRANSPORT_FD_OFF, /* to transfer file descriptors */
 
 	SERVICE_FD_MAX
 };
@@ -31,5 +32,6 @@ extern int install_service_fd(enum sfd_type type, int fd);
 extern int close_service_fd(enum sfd_type type);
 extern bool is_service_fd(int fd, enum sfd_type type);
 extern bool is_any_service_fd(int fd);
+extern int service_fd_min_fd(void);
 
 #endif /* __CR_SERVICE_FD_H__ */
