@@ -10,8 +10,13 @@
 
 struct rt_sigframe;
 
+#ifndef SIGFRAME_MAX_OFFSET
+#define SIGFRAME_MAX_OFFSET RT_SIGFRAME_OFFSET(0)
+#endif
+
 /* sigframe should be aligned on 64 byte for x86 and 8 bytes for arm */
-#define RESTORE_STACK_SIGFRAME ALIGN(sizeof(struct rt_sigframe) + SIGFRAME_OFFSET, 64)
+#define RESTORE_STACK_SIGFRAME						\
+	ALIGN(sizeof(struct rt_sigframe) + SIGFRAME_MAX_OFFSET, 64)
 
 #ifndef __ARCH_SI_PREAMBLE_SIZE
 #define __ARCH_SI_PREAMBLE_SIZE	(3 * sizeof(int))
@@ -49,18 +54,4 @@ extern int construct_sigframe(struct rt_sigframe *sigframe,
 			      struct rt_sigframe *rsigframe,
 			      CoreEntry *core);
 
-/*
- * FIXME Convert it to inline helper, which requires
- *	 to unweave types mess we've generated for
- *	 run-time data.
- */
-#define setup_sas(sigframe, sas)											\
-do {															\
-	if ((sas)) {													\
-		RT_SIGFRAME_UC((sigframe)).uc_stack.ss_sp	= (void *)decode_pointer((sas)->ss_sp);			\
-		RT_SIGFRAME_UC((sigframe)).uc_stack.ss_flags	= (int)(sas)->ss_flags;					\
-		RT_SIGFRAME_UC((sigframe)).uc_stack.ss_size	= (size_t)(sas)->ss_size;				\
-	}														\
-} while (0)
-
 #endif /* __CR_SIGFRAME_H__ */
