@@ -9,6 +9,10 @@
 /* Kernel's limit */
 #define TERMIOS_NCC	19
 
+/* Popular serial console's majors, which not defined in <linux/major.h> */
+#define USB_SERIAL_MAJOR	188
+#define LOW_DENSE_SERIAL_MAJOR	204
+
 extern const struct fdtype_ops tty_dump_ops;
 
 struct tty_driver;
@@ -18,13 +22,12 @@ static inline int is_tty(dev_t rdev, dev_t dev)
 	return get_tty_driver(rdev, dev) != NULL;
 }
 
+extern int tty_post_actions(void);
 extern int dump_verify_tty_sids(void);
 extern struct collect_image_info tty_info_cinfo;
 extern struct collect_image_info tty_cinfo;
+extern struct collect_image_info tty_cdata;
 extern int prepare_shared_tty(void);
-extern int tty_setup_slavery(void);
-
-extern int tty_verify_active_pairs(void);
 
 extern int tty_prep_fds(void);
 extern void tty_fini_fds(void);
