@@ -10,8 +10,12 @@
 #include <signal.h>
 #include <sched.h>
 #include <sys/capability.h>
+#include <sys/stat.h>
+#include <limits.h>
+#include <errno.h>
 
-#include "util.h"
+#include "rst-malloc.h"
+#include "cr_options.h"
 #include "imgset.h"
 #include "uts_ns.h"
 #include "ipc_ns.h"
@@ -35,6 +39,149 @@ static struct ns_desc *ns_desc_array[] = {
 	&cgroup_ns_desc,
 };
 
+static unsigned int join_ns_flags;
+
+int check_namespace_opts(void)
+{
+	errno = 22;
+	if (join_ns_flags & opts.empty_ns) {
+		pr_perror("Conflict flags: -join-ns and -empty-ns");
+		return -1;
+	}
+	if (join_ns_flags & CLONE_NEWUSER)
+		pr_warn("join-ns with user-namespace is not fully tested and dangerous");
+
+	errno = 0;
+	return 0;
+}
+
+static int check_int_str(char *str)
+{
+	char *endptr;
+	long val;
+
+	if (str == NULL)
+		return 0;
+
+	if (*str == '\0') {
+		str = NULL;
+		return 0;
+	}
+
+	errno = 22;
+	val = strtol(str, &endptr, 10);
+	if ((errno == ERANGE) || (endptr == str)
+			|| (*endptr != '\0')
+			|| (val < 0) || (val > 65535)) {
+		str = NULL;
+		return -1;
+	}
+
+	errno = 0;
+	return 0;
+}
+
+static int check_ns_file(char *ns_file)
+{
+	int pid, ret, proc_dir;
+
+	if (!check_int_str(ns_file)) {
+		pid = atoi(ns_file);
+		if (pid <= 0) {
+			pr_perror("Invalid join_ns pid %s", ns_file);
+			return -1;
+		}
+		proc_dir = open_pid_proc(pid);
+		if (proc_dir < 0) {
+			pr_perror("Invalid join_ns pid: /proc/%s not found", ns_file);
+			return -1;
+		}
+		return 0;
+	}
+
+	ret = access(ns_file, 0);
+	if (ret < 0) {
+		pr_perror("Can't access join-ns file: %s", ns_file);
+		return -1;
+	}
+	return 0;
+}
+
+static int set_user_extra_opts(struct join_ns *jn, char *extra_opts)
+{
+	char *uid, *gid, *aux;
+
+	if (extra_opts == NULL) {
+		jn->extra_opts.user_extra.uid = NULL;
+		jn->extra_opts.user_extra.gid = NULL;
+		return 0;
+	}
+
+	uid = extra_opts;
+	aux = strchr(extra_opts, ',');
+	if (aux == NULL) {
+		gid = NULL;
+	} else {
+		*aux = '\0';
+		gid = aux + 1;
+	}
+
+	if (check_int_str(uid) || check_int_str(gid))
+		return -1;
+
+	jn->extra_opts.user_extra.uid = uid;
+	jn->extra_opts.user_extra.gid = gid;
+
+	return 0;
+}
+
+int join_ns_add(const char *type, char *ns_file, char *extra_opts)
+{
+	struct join_ns *jn;
+
+	if (check_ns_file(ns_file))
+		return -1;
+
+	jn = xmalloc(sizeof(*jn));
+	if (!jn)
+		return -1;
+
+	jn->ns_file = ns_file;
+	if (!strncmp(type, "net", 4)) {
+		jn->nd = &net_ns_desc;
+		join_ns_flags |= CLONE_NEWNET;
+	} else if (!strncmp(type, "uts", 4)) {
+		jn->nd = &uts_ns_desc;
+		join_ns_flags |= CLONE_NEWUTS;
+	} else if (!strncmp(type, "ipc", 4)) {
+		jn->nd = &ipc_ns_desc;
+		join_ns_flags |= CLONE_NEWIPC;
+	} else if (!strncmp(type, "pid", 4)) {
+		pr_err("join-ns pid namespace not supported\n");
+		goto err;
+	} else if (!strncmp(type, "user", 5)) {
+		jn->nd = &user_ns_desc;
+		if (set_user_extra_opts(jn, extra_opts)) {
+			pr_err("invalid user namespace extra_opts %s\n", extra_opts);
+			goto err;
+		}
+		join_ns_flags |= CLONE_NEWUSER;
+	} else if (!strncmp(type, "mnt", 4)) {
+		jn->nd = &mnt_ns_desc;
+		join_ns_flags |= CLONE_NEWNS;
+	} else {
+		pr_err("invalid namespace type %s\n", type);
+		goto err;
+	}
+
+	list_add_tail(&jn->list, &opts.join_ns);
+	pr_info("Added %s:%s join namespace\n", type, ns_file);
+	return 0;
+err:
+	xfree(jn);
+	return -1;
+}
+
 static unsigned int parse_ns_link(char *link, size_t len, struct ns_desc *d)
 {
 	unsigned long kid = 0;
@@ -78,7 +225,7 @@ int switch_ns(int pid, struct ns_desc *nd, int *rst)
 
 	nsfd = open_proc(pid, "ns/%s", nd->str);
 	if (nsfd < 0) {
-		pr_perror("Can't open ipcns file");
+		pr_perror("Can't open ns file");
 		goto err_ns;
 	}
 
@@ -146,7 +293,7 @@ struct ns_id *rst_new_ns_id(unsigned int id, pid_t pid,
 	if (nsid) {
 		nsid->type = type;
 		nsid_add(nsid, nd, id, pid);
-		futex_set(&nsid->ns_populated, 0);
+		nsid->ns_populated = false;
 	}
 
 	return nsid;
@@ -259,13 +406,13 @@ static unsigned int generate_ns_id(int pid, unsigned int kid, struct ns_desc *nd
 	} else
 		type = NS_CRIU;
 
-	nsid = xmalloc(sizeof(*nsid));
+	nsid = xzalloc(sizeof(*nsid));
 	if (!nsid)
 		return 0;
 
 	nsid->type = type;
 	nsid->kid = kid;
-	futex_set(&nsid->ns_populated, 1);
+	nsid->ns_populated = true;
 	nsid_add(nsid, nd, ns_next_id++, pid);
 
 found:
@@ -276,28 +423,27 @@ found:
 
 static unsigned int __get_ns_id(int pid, struct ns_desc *nd, protobuf_c_boolean *supported, struct ns_id **ns)
 {
-	int proc_dir, ret;
+	int proc_dir;
 	unsigned int kid;
-	char ns_path[10], ns_id[32];
+	char ns_path[10];
+	struct stat st;
 
 	proc_dir = open_pid_proc(pid);
 	if (proc_dir < 0)
 		return 0;
 
 	sprintf(ns_path, "ns/%s", nd->str);
-	ret = readlinkat(proc_dir, ns_path, ns_id, sizeof(ns_id) - 1);
-	if (ret < 0) {
+
+	if (fstatat(proc_dir, ns_path, &st, 0)) {
 		if (errno == ENOENT) {
 			/* The namespace is unsupported */
 			kid = 0;
 			goto out;
 		}
-		pr_perror("Can't readlink ns link");
-		return 0;
+		pr_perror("Unable to stat %s", ns_path);
+		return -1;
 	}
-	ns_id[ret] = '\0';
-
-	kid = parse_ns_link(ns_id, ret, nd);
+	kid = st.st_ino;
 	BUG_ON(!kid);
 
 out:
@@ -412,7 +558,7 @@ static struct file_desc_ops ns_desc_ops = {
 	.open = open_ns_fd,
 };
 
-static int collect_one_nsfile(void *o, ProtobufCMessage *base)
+static int collect_one_nsfile(void *o, ProtobufCMessage *base, struct cr_img *img)
 {
 	struct ns_file_info *nfi = o;
 
@@ -650,7 +796,7 @@ int collect_user_namespaces(bool for_dump)
 	if (!(root_ns_mask & CLONE_NEWUSER))
 		return 0;
 
-	return walk_namespaces(&net_ns_desc, collect_user_ns, NULL);
+	return walk_namespaces(&user_ns_desc, collect_user_ns, NULL);
 }
 
 static int check_user_ns(int pid)
@@ -689,7 +835,7 @@ static int check_user_ns(int pid)
 		}
 
 		if (setgroups(0, NULL) < 0) {
-			pr_perror("Unable to drop supplementary groups\n");
+			pr_perror("Unable to drop supplementary groups");
 			return -1;
 		}
 
@@ -829,11 +975,6 @@ static int do_dump_namespaces(struct ns_id *ns)
 				ns->id, ns->ns_pid);
 		ret = dump_net_ns(ns->id);
 		break;
-	case CLONE_NEWCGROUP:
-		pr_info("Dump CGROUP namespace info %d via %d\n",
-				ns->id, ns->ns_pid);
-		/* handled separately in cgroup dumping code */
-		break;
 	default:
 		pr_err("Unknown namespace flag %x\n", ns->nd->cflag);
 		break;
@@ -880,6 +1021,8 @@ int dump_namespaces(struct pstree_item *item, unsigned int ns_flags)
 			case CLONE_NEWNS:
 			/* Userns is dumped before dumping tasks */
 			case CLONE_NEWUSER:
+			/* handled separately in cgroup dumping code */
+			case CLONE_NEWCGROUP:
 				continue;
 		}
 
@@ -952,7 +1095,7 @@ struct unsc_msg {
 	/*
 	 * 0th is the call address
 	 * 1st is the flags
-	 * 2nd is the optional (NULL in responce) arguments
+	 * 2nd is the optional (NULL in response) arguments
 	 */
 	struct iovec iov[3];
 	char c[CMSG_SPACE(sizeof(struct ucred)) + CMSG_SPACE(sizeof(int))];
@@ -1132,10 +1275,10 @@ int __userns_call(const char *func_name, uns_call_t call, int flags,
 		/*
 		 * Why don't we lock for async requests? Because
 		 * they just put the request in the daemon's
-		 * queue and do not wait for the responce. Thus
-		 * when daemon responce there's only one client
+		 * queue and do not wait for the response. Thus
+		 * when daemon response there's only one client
 		 * waiting for it in recvmsg below, so he
-		 * responces to proper caller.
+		 * responses to proper caller.
 		 */
 		mutex_lock(&task_entries->userns_sync_lock);
 	else
@@ -1201,7 +1344,7 @@ static int start_usernsd(void)
 	 * b) Make callers note the damon death by seeing the
 	 *    disconnected socket. In case of dgram socket
 	 *    callers would just get stuck in receiving the
-	 *    responce.
+	 *    response.
 	 */
 
 	if (socketpair(PF_UNIX, SOCK_SEQPACKET, 0, sk)) {
@@ -1371,6 +1514,126 @@ static int prepare_userns_creds()
 	return 0;
 }
 
+static int get_join_ns_fd(struct join_ns *jn)
+{
+	int pid, fd;
+	char nsf[32];
+	char *pnsf;
+
+	pid = atoi(jn->ns_file);
+	if (pid > 0) {
+		snprintf(nsf, sizeof(nsf), "/proc/%d/ns/%s", pid, jn->nd->str);
+		pnsf = nsf;
+	} else {
+		pnsf = jn->ns_file;
+	}
+
+	fd = open(pnsf, O_RDONLY);
+	if (fd < 0) {
+		pr_perror("Can't open ns file: %s", pnsf);
+		return -1;
+	}
+	jn->ns_fd = fd;
+	return 0;
+}
+
+static int switch_join_ns(struct join_ns *jn)
+{
+	struct stat st, self_st;
+	char buf[32];
+
+	if (jn->nd == &user_ns_desc) {
+		/* It is not permitted to use setns() to reenter the caller's current
+		 * user namespace.  This prevents a caller that has dropped capabilities
+		 * from regaining those capabilities via a call to setns()
+		 */
+		if (fstat(jn->ns_fd, &st) == -1) {
+			pr_perror("Can't get ns file %s stat", jn->ns_file);
+			return -1;
+		}
+
+		snprintf(buf, sizeof(buf), "/proc/self/ns/%s", jn->nd->str);
+		if (stat(buf, &self_st) == -1) {
+			pr_perror("Can't get ns file %s stat", buf);
+			return -1;
+		}
+
+		if (st.st_ino == self_st.st_ino)
+			return 0;
+	}
+
+	if (setns(jn->ns_fd, jn->nd->cflag)) {
+		pr_perror("Failed to setns when join-ns %s:%s", jn->nd->str, jn->ns_file);
+		return -1;
+	}
+
+	return 0;
+}
+
+static int switch_user_join_ns(struct join_ns *jn)
+{
+	uid_t uid;
+	gid_t gid;
+
+	if (jn == NULL)
+		return 0;
+
+	if (switch_join_ns(jn))
+		return -1;
+
+	if (jn->extra_opts.user_extra.uid == NULL)
+		uid = getuid();
+	else
+		uid = atoi(jn->extra_opts.user_extra.uid);
+
+	if (jn->extra_opts.user_extra.gid == NULL)
+		gid = getgid();
+	else
+		gid = atoi(jn->extra_opts.user_extra.gid);
+
+	/* FIXME:
+	 * if err occurs in setuid/setgid, should we just alert or
+	 * return an error
+	 */
+	if (setuid(uid)) {
+		pr_perror("setuid failed while joining userns");
+		return -1;
+	}
+	if (setgid(gid)) {
+		pr_perror("setgid failed while joining userns");
+		return -1;
+	}
+
+	return 0;
+}
+
+int join_namespaces(void)
+{
+	struct join_ns *jn, *user_jn = NULL;
+	int ret = -1;
+
+	list_for_each_entry(jn, &opts.join_ns, list)
+		if (get_join_ns_fd(jn))
+			goto err_out;
+
+	list_for_each_entry(jn, &opts.join_ns, list)
+		if (jn->nd == &user_ns_desc) {
+			user_jn = jn;
+		} else {
+			if (switch_join_ns(jn))
+				goto err_out;
+		}
+
+	if (switch_user_join_ns(user_jn))
+		goto err_out;
+
+	ret = 0;
+err_out:
+	list_for_each_entry(jn, &opts.join_ns, list)
+		close_safe(&jn->ns_fd);
+	return ret;
+}
+
 int prepare_namespace(struct pstree_item *item, unsigned long clone_flags)
 {
 	pid_t pid = item->pid.virt;
