@@ -61,7 +61,23 @@ int read_ns_sys_file(char *path, char *buf, int len)
 	return rlen;
 }
 
-static char *devconfs[] = {
+static bool sysctl_entries_equal(SysctlEntry *a, SysctlEntry *b)
+{
+	if (a->type != b->type)
+		return false;
+
+	switch (a->type) {
+		case SYSCTL_TYPE__CTL_32:
+			return a->has_iarg && b->has_iarg && a->iarg == b->iarg;
+		case SYSCTL_TYPE__CTL_STR:
+			return a->sarg && b->sarg && !strcmp(a->sarg, b->sarg);
+		default:;
+	}
+
+	return false;
+}
+
+static char *devconfs4[] = {
 	"accept_local",
 	"accept_redirects",
 	"accept_source_route",
@@ -90,8 +106,185 @@ static char *devconfs[] = {
 	"src_valid_mark",
 	"tag",
 	"ignore_routes_with_linkdown",
+	"drop_gratuitous_arp",
+	"drop_unicast_in_l2_multicast",
 };
 
+char *devconfs6[] = {
+	"accept_dad",
+	"accept_ra",
+	"accept_ra_defrtr",
+	"accept_ra_from_local",
+	"accept_ra_min_hop_limit",
+	"accept_ra_mtu",
+	"accept_ra_pinfo",
+	"accept_ra_rt_info_max_plen",
+	"accept_ra_rtr_pref",
+	"accept_redirects",
+	"accept_source_route",
+	"autoconf",
+	"dad_transmits",
+	"disable_ipv6",
+	"drop_unicast_in_l2_multicast",
+	"drop_unsolicited_na",
+	"force_mld_version",
+	"force_tllao",
+	"forwarding",
+	"hop_limit",
+	"ignore_routes_with_linkdown",
+	"keep_addr_on_down",
+	"max_addresses",
+	"max_desync_factor",
+	"mldv1_unsolicited_report_interval",
+	"mldv2_unsolicited_report_interval",
+	"mtu",
+	"ndisc_notify",
+	"optimistic_dad",
+	"proxy_ndp",
+	"regen_max_retry",
+	"router_probe_interval",
+	"router_solicitation_delay",
+	"router_solicitation_interval",
+	"router_solicitations",
+	"stable_secret",
+	"suppress_frag_ndisc",
+	"temp_prefered_lft",
+	"temp_valid_lft",
+	"use_oif_addrs_only",
+	"use_optimistic",
+	"use_tempaddr",
+};
+
+#define CONF_OPT_PATH "net/%s/conf/%s/%s"
+#define MAX_CONF_OPT_PATH IFNAMSIZ+60
+#define MAX_STR_CONF_LEN 200
+
+static int net_conf_op(char *tgt, SysctlEntry **conf, int n, int op, char *proto,
+		struct sysctl_req *req, char (*path)[MAX_CONF_OPT_PATH], int size,
+		char **devconfs, SysctlEntry **def_conf)
+{
+	int i, ri, ar = -1;
+	int ret, flags = op == CTL_READ ? CTL_FLAGS_OPTIONAL : 0;
+	SysctlEntry **rconf;
+
+	if (n > size)
+		pr_warn("The image contains unknown sysctl-s\n");
+
+	rconf = xmalloc(sizeof(SysctlEntry *) * size);
+	if (!rconf)
+		return -1;
+
+	for (i = 0, ri = 0; i < size; i++) {
+		if (i >= n) {
+			pr_warn("Skip %s/%s\n", tgt, devconfs[i]);
+			continue;
+		}
+		/*
+		 * If dev conf value is the same as default skip restoring it,
+		 * mtu may be changed by disable_ipv6 so we can not skip
+		 * it's restore
+		 */
+		if (def_conf && sysctl_entries_equal(conf[i], def_conf[i])
+				&& strcmp(devconfs[i], "mtu")) {
+			pr_debug("Skip %s/%s, coincides with default\n", tgt, devconfs[i]);
+			continue;
+		}
+
+		/*
+		 * Make "accept_redirects" go last on write(it should
+		 * restore after forwarding to be correct)
+		 */
+		if (op == CTL_WRITE && !strcmp(devconfs[i], "accept_redirects")) {
+			ar = i;
+			continue;
+		}
+
+		snprintf(path[i], MAX_CONF_OPT_PATH, CONF_OPT_PATH, proto, tgt, devconfs[i]);
+		req[ri].name = path[i];
+		switch (conf[i]->type) {
+			case SYSCTL_TYPE__CTL_32:
+				req[ri].type = CTL_32;
+
+				/* skip non-existing sysctl */
+				if (op == CTL_WRITE && !conf[i]->has_iarg)
+					continue;
+
+				req[ri].arg = &conf[i]->iarg;
+				break;
+			case SYSCTL_TYPE__CTL_STR:
+				req[ri].type = CTL_STR(MAX_STR_CONF_LEN);
+				flags |= op == CTL_READ && !strcmp(devconfs[i], "stable_secret")
+					? CTL_FLAGS_READ_EIO_SKIP : 0;
+
+				/* skip non-existing sysctl */
+				if (op == CTL_WRITE && !conf[i]->sarg)
+					continue;
+
+				req[ri].arg = conf[i]->sarg;
+				break;
+			default:
+				continue;
+		}
+		req[ri].flags = flags;
+		rconf[ri] = conf[i];
+		ri++;
+	}
+
+	if (ar != -1
+	    && conf[ar]->type == SYSCTL_TYPE__CTL_32
+	    && conf[ar]->has_iarg) {
+		snprintf(path[ar], MAX_CONF_OPT_PATH, CONF_OPT_PATH, proto, tgt, devconfs[ar]);
+		req[ri].name = path[ar];
+		req[ri].type = CTL_32;
+		req[ri].arg = &conf[ar]->iarg;
+		req[ri].flags = flags;
+		rconf[ri] = conf[ar];
+		ri++;
+	}
+
+	ret = sysctl_op(req, ri, op, CLONE_NEWNET);
+	if (ret < 0) {
+		pr_err("Failed to %s %s/<confs>\n", (op == CTL_READ)?"read":"write", tgt);
+		goto err_free;
+	}
+
+	if (op == CTL_READ) {
+		/* (un)mark (non-)existing sysctls in image */
+		for (i = 0; i < ri; i++)
+			if (req[i].flags & CTL_FLAGS_HAS) {
+				if (rconf[i]->type == SYSCTL_TYPE__CTL_32)
+					rconf[i]->has_iarg = true;
+			} else {
+				if (rconf[i]->type == SYSCTL_TYPE__CTL_STR)
+					rconf[i]->sarg = NULL;
+			}
+	}
+
+err_free:
+	xfree(rconf);
+	return ret;
+}
+
+static int ipv4_conf_op(char *tgt, SysctlEntry **conf, int n, int op, SysctlEntry **def_conf)
+{
+	struct sysctl_req req[ARRAY_SIZE(devconfs4)];
+	char path[ARRAY_SIZE(devconfs4)][MAX_CONF_OPT_PATH];
+
+	return net_conf_op(tgt, conf, n, op, "ipv4",
+			req, path, ARRAY_SIZE(devconfs4),
+			devconfs4, def_conf);
+}
+
+static int ipv6_conf_op(char *tgt, SysctlEntry **conf, int n, int op, SysctlEntry **def_conf)
+{
+	struct sysctl_req req[ARRAY_SIZE(devconfs6)];
+	char path[ARRAY_SIZE(devconfs6)][MAX_CONF_OPT_PATH];
+
+	return net_conf_op(tgt, conf, n, op, "ipv6",
+			req, path, ARRAY_SIZE(devconfs6),
+			devconfs6, def_conf);
+}
+
 /*
  * I case if some entry is missing in
  * the kernel, simply write DEVCONFS_UNUSED
@@ -99,29 +292,26 @@ static char *devconfs[] = {
  */
 #define DEVCONFS_UNUSED        (-1u)
 
-#define NET_CONF_PATH "net/ipv4/conf"
-#define MAX_CONF_OPT_PATH IFNAMSIZ+50
-
-static int ipv4_conf_op(char *tgt, int *conf, int n, int op, NetnsEntry **netns)
+static int ipv4_conf_op_old(char *tgt, int *conf, int n, int op, int *def_conf)
 {
 	int i, ri;
 	int ret, flags = op == CTL_READ ? CTL_FLAGS_OPTIONAL : 0;
-	struct sysctl_req req[ARRAY_SIZE(devconfs)];
-	char path[ARRAY_SIZE(devconfs)][MAX_CONF_OPT_PATH];
+	struct sysctl_req req[ARRAY_SIZE(devconfs4)];
+	char path[ARRAY_SIZE(devconfs4)][MAX_CONF_OPT_PATH];
 
-	if (n > ARRAY_SIZE(devconfs))
+	if (n > ARRAY_SIZE(devconfs4))
 		pr_warn("The image contains unknown sysctl-s\n");
 
-	for (i = 0, ri = 0; i < ARRAY_SIZE(devconfs); i++) {
+	for (i = 0, ri = 0; i < ARRAY_SIZE(devconfs4); i++) {
 		if (i >= n) {
-			pr_warn("Skip %s/%s\n", tgt, devconfs[i]);
+			pr_warn("Skip %s/%s\n", tgt, devconfs4[i]);
 			continue;
 		}
 		/*
 		 * If dev conf value is the same as default skip restoring it
 		 */
-		if (netns && conf[i] == (*netns)->def_conf[i]) {
-			pr_debug("DEBUG Skip %s/%s, val =%d\n", tgt, devconfs[i], conf[i]);
+		if (def_conf && conf[i] == def_conf[i]) {
+			pr_debug("DEBUG Skip %s/%s, val =%d\n", tgt, devconfs4[i], conf[i]);
 			continue;
 		}
 
@@ -130,7 +320,7 @@ static int ipv4_conf_op(char *tgt, int *conf, int n, int op, NetnsEntry **netns)
 		else if (op == CTL_READ)
 			conf[i] = DEVCONFS_UNUSED;
 
-		snprintf(path[i], MAX_CONF_OPT_PATH, "%s/%s/%s", NET_CONF_PATH, tgt, devconfs[i]);
+		snprintf(path[i], MAX_CONF_OPT_PATH, CONF_OPT_PATH, "ipv4", tgt, devconfs4[i]);
 		req[ri].name = path[i];
 		req[ri].arg = &conf[i];
 		req[ri].type = CTL_32;
@@ -152,11 +342,17 @@ int write_netdev_img(NetDeviceEntry *nde, struct cr_imgset *fds)
 }
 
 static int dump_one_netdev(int type, struct ifinfomsg *ifi,
-		struct rtattr **tb, struct cr_imgset *fds,
+		struct nlattr **tb, struct cr_imgset *fds,
 		int (*dump)(NetDeviceEntry *, struct cr_imgset *))
 {
-	int ret;
+	int ret = -1;
+	int i;
 	NetDeviceEntry netdev = NET_DEVICE_ENTRY__INIT;
+	SysctlEntry *confs4 = NULL;
+	int size4 = ARRAY_SIZE(devconfs4);
+	SysctlEntry *confs6 = NULL;
+	int size6 = ARRAY_SIZE(devconfs6);
+	char stable_secret[MAX_STR_CONF_LEN + 1] = {};
 
 	if (!tb[IFLA_IFNAME]) {
 		pr_err("No name for link %d\n", ifi->ifi_index);
@@ -171,19 +367,53 @@ static int dump_one_netdev(int type, struct ifinfomsg *ifi,
 
 	if (tb[IFLA_ADDRESS] && (type != ND_TYPE__LOOPBACK)) {
 		netdev.has_address = true;
-		netdev.address.data = RTA_DATA(tb[IFLA_ADDRESS]);
-		netdev.address.len = RTA_PAYLOAD(tb[IFLA_ADDRESS]);
+		netdev.address.data = nla_data(tb[IFLA_ADDRESS]);
+		netdev.address.len = nla_len(tb[IFLA_ADDRESS]);
 		pr_info("Found ll addr (%02x:../%d) for %s\n",
 				(int)netdev.address.data[0],
 				(int)netdev.address.len, netdev.name);
 	}
 
-	netdev.n_conf = ARRAY_SIZE(devconfs);
-	netdev.conf = xmalloc(sizeof(int) * netdev.n_conf);
-	if (!netdev.conf)
-		return -1;
+	netdev.n_conf4 = size4;
+	netdev.conf4 = xmalloc(sizeof(SysctlEntry *) * size4);
+	if (!netdev.conf4)
+		goto err_free;
 
-	ret = ipv4_conf_op(netdev.name, netdev.conf, netdev.n_conf, CTL_READ, NULL);
+	confs4 = xmalloc(sizeof(SysctlEntry) * size4);
+	if (!confs4)
+		goto err_free;
+
+	for (i = 0; i < size4; i++) {
+		sysctl_entry__init(&confs4[i]);
+		netdev.conf4[i] = &confs4[i];
+		netdev.conf4[i]->type = CTL_32;
+	}
+
+	netdev.n_conf6 = size6;
+	netdev.conf6 = xmalloc(sizeof(SysctlEntry *) * size6);
+	if (!netdev.conf6)
+		goto err_free;
+
+	confs6 = xmalloc(sizeof(SysctlEntry) * size6);
+	if (!confs6)
+		goto err_free;
+
+	for (i = 0; i < size6; i++) {
+		sysctl_entry__init(&confs6[i]);
+		netdev.conf6[i] = &confs6[i];
+		if (strcmp(devconfs6[i], "stable_secret")) {
+			netdev.conf6[i]->type = SYSCTL_TYPE__CTL_32;
+		} else {
+			netdev.conf6[i]->type = SYSCTL_TYPE__CTL_STR;
+			netdev.conf6[i]->sarg = stable_secret;
+		}
+	}
+
+	ret = ipv4_conf_op(netdev.name, netdev.conf4, size4, CTL_READ, NULL);
+	if (ret < 0)
+		goto err_free;
+
+	ret = ipv6_conf_op(netdev.name, netdev.conf6, size6, CTL_READ, NULL);
 	if (ret < 0)
 		goto err_free;
 
@@ -192,30 +422,33 @@ static int dump_one_netdev(int type, struct ifinfomsg *ifi,
 
 	ret = dump(&netdev, fds);
 err_free:
-	xfree(netdev.conf);
+	xfree(netdev.conf4);
+	xfree(confs4);
+	xfree(netdev.conf6);
+	xfree(confs6);
 	return ret;
 }
 
-static char *link_kind(struct ifinfomsg *ifi, struct rtattr **tb)
+static char *link_kind(struct ifinfomsg *ifi, struct nlattr **tb)
 {
-	struct rtattr *linkinfo[IFLA_INFO_MAX + 1];
+	struct nlattr *linkinfo[IFLA_INFO_MAX + 1];
 
 	if (!tb[IFLA_LINKINFO]) {
 		pr_err("No linkinfo for eth link %d\n", ifi->ifi_index);
 		return NULL;
 	}
 
-	parse_rtattr_nested(linkinfo, IFLA_INFO_MAX, tb[IFLA_LINKINFO]);
+	nla_parse_nested(linkinfo, IFLA_INFO_MAX, tb[IFLA_LINKINFO], NULL);
 	if (!linkinfo[IFLA_INFO_KIND]) {
 		pr_err("No kind for eth link %d\n", ifi->ifi_index);
 		return NULL;
 	}
 
-	return RTA_DATA(linkinfo[IFLA_INFO_KIND]);
+	return nla_data(linkinfo[IFLA_INFO_KIND]);
 }
 
 static int dump_unknown_device(struct ifinfomsg *ifi, char *kind,
-		struct rtattr **tb, struct cr_imgset *fds)
+		struct nlattr **tb, struct cr_imgset *fds)
 {
 	int ret;
 
@@ -265,7 +498,7 @@ static int dump_bridge(NetDeviceEntry *nde, struct cr_imgset *imgset)
 }
 
 static int dump_one_ethernet(struct ifinfomsg *ifi, char *kind,
-		struct rtattr **tb, struct cr_imgset *fds)
+		struct nlattr **tb, struct cr_imgset *fds)
 {
 	if (!strcmp(kind, "veth"))
 		/*
@@ -281,12 +514,27 @@ static int dump_one_ethernet(struct ifinfomsg *ifi, char *kind,
 		return dump_one_netdev(ND_TYPE__TUN, ifi, tb, fds, dump_tun_link);
 	if (!strcmp(kind, "bridge"))
 		return dump_one_netdev(ND_TYPE__BRIDGE, ifi, tb, fds, dump_bridge);
+	if (!strcmp(kind, "gretap")) {
+		char *name = (char *)RTA_DATA(tb[IFLA_IFNAME]);
+
+		if (!name) {
+			pr_err("gretap %d has no name\n", ifi->ifi_index);
+			return -1;
+		}
+
+		if (!strcmp(name, "gretap0")) {
+			pr_info("found %s, ignoring\n", name);
+			return 0;
+		}
+
+		pr_warn("GRE tap device %s not supported natively\n", name);
+	}
 
 	return dump_unknown_device(ifi, kind, tb, fds);
 }
 
 static int dump_one_gendev(struct ifinfomsg *ifi, char *kind,
-		struct rtattr **tb, struct cr_imgset *fds)
+		struct nlattr **tb, struct cr_imgset *fds)
 {
 	if (!strcmp(kind, "tun"))
 		return dump_one_netdev(ND_TYPE__TUN, ifi, tb, fds, dump_tun_link);
@@ -295,7 +543,7 @@ static int dump_one_gendev(struct ifinfomsg *ifi, char *kind,
 }
 
 static int dump_one_voiddev(struct ifinfomsg *ifi, char *kind,
-		struct rtattr **tb, struct cr_imgset *fds)
+		struct nlattr **tb, struct cr_imgset *fds)
 {
 	if (!strcmp(kind, "venet"))
 		return dump_one_netdev(ND_TYPE__VENET, ifi, tb, fds, NULL);
@@ -303,12 +551,33 @@ static int dump_one_voiddev(struct ifinfomsg *ifi, char *kind,
 	return dump_unknown_device(ifi, kind, tb, fds);
 }
 
+static int dump_one_gre(struct ifinfomsg *ifi, char *kind,
+		struct nlattr **tb, struct cr_imgset *fds)
+{
+	if (!strcmp(kind, "gre")) {
+		char *name = (char *)RTA_DATA(tb[IFLA_IFNAME]);
+		if (!name) {
+			pr_err("gre device %d has no name\n", ifi->ifi_index);
+			return -1;
+		}
+
+		if (!strcmp(name, "gre0")) {
+			pr_info("found %s, ignoring\n", name);
+			return 0;
+		}
+
+		pr_warn("GRE tunnel device %s not supported natively\n", name);
+	}
+
+	return dump_unknown_device(ifi, kind, tb, fds);
+}
+
 static int dump_one_link(struct nlmsghdr *hdr, void *arg)
 {
 	struct cr_imgset *fds = arg;
 	struct ifinfomsg *ifi;
 	int ret = 0, len = hdr->nlmsg_len - NLMSG_LENGTH(sizeof(*ifi));
-	struct rtattr *tb[IFLA_MAX + 1];
+	struct nlattr *tb[IFLA_MAX + 1];
 	char *kind;
 
 	ifi = NLMSG_DATA(hdr);
@@ -318,7 +587,7 @@ static int dump_one_link(struct nlmsghdr *hdr, void *arg)
 		return -1;
 	}
 
-	parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), len);
+	nlmsg_parse(hdr, sizeof(struct ifinfomsg), tb, IFLA_MAX, NULL);
 	pr_info("\tLD: Got link %d, type %d\n", ifi->ifi_index, ifi->ifi_type);
 
 	if (ifi->ifi_type == ARPHRD_LOOPBACK) 
@@ -338,6 +607,9 @@ static int dump_one_link(struct nlmsghdr *hdr, void *arg)
 	case ARPHRD_VOID:
 		ret = dump_one_voiddev(ifi, kind, tb, fds);
 		break;
+	case ARPHRD_IPGRE:
+		ret = dump_one_gre(ifi, kind, tb, fds);
+		break;
 	default:
 unk:
 		ret = dump_unknown_device(ifi, kind, tb, fds);
@@ -731,6 +1003,8 @@ static int restore_links(int pid, NetnsEntry **netns)
 	}
 
 	while (1) {
+		NetnsEntry **def_netns = netns;
+
 		ret = pb_read_one_eof(img, &nde, PB_NETDEV);
 		if (ret <= 0)
 			break;
@@ -741,17 +1015,23 @@ static int restore_links(int pid, NetnsEntry **netns)
 			goto exit;
 		}
 
-		if (nde->conf) {
-			NetnsEntry **def_netns = netns;
-			/*
-			 * optimize restore of devices configuration except lo
-			 * lo is created with namespace and before default is set
-			 * so we cant optimize its restore
-			 */
-			if (nde->type == ND_TYPE__LOOPBACK)
-				def_netns = NULL;
-			ret = ipv4_conf_op(nde->name, nde->conf, nde->n_conf, CTL_WRITE, def_netns);
-		}
+		/*
+		 * optimize restore of devices configuration except lo
+		 * lo is created with namespace and before default is set
+		 * so we can't optimize its restore
+		 */
+		if (nde->type == ND_TYPE__LOOPBACK)
+			def_netns = NULL;
+
+		if (nde->conf4)
+			ret = ipv4_conf_op(nde->name, nde->conf4, nde->n_conf4, CTL_WRITE, def_netns ? (*def_netns)->def_conf4 : NULL);
+		else if (nde->conf)
+			ret = ipv4_conf_op_old(nde->name, nde->conf, nde->n_conf, CTL_WRITE, def_netns ? (*def_netns)->def_conf : NULL);
+		if (ret)
+			goto exit;
+
+		if (nde->conf6)
+			ret = ipv6_conf_op(nde->name, nde->conf6, nde->n_conf6, CTL_WRITE, def_netns ? (*def_netns)->def_conf6 : NULL);
 exit:
 		net_device_entry__free_unpacked(nde, NULL);
 		if (ret)
@@ -866,32 +1146,95 @@ static inline int dump_iptables(struct cr_imgset *fds)
 
 static int dump_netns_conf(struct cr_imgset *fds)
 {
-	int ret, n;
+	int ret = -1;
+	int i;
 	NetnsEntry netns = NETNS_ENTRY__INIT;
+	SysctlEntry *def_confs4 = NULL, *all_confs4 = NULL;
+	int size4 = ARRAY_SIZE(devconfs4);
+	SysctlEntry *def_confs6 = NULL, *all_confs6 = NULL;
+	int size6 = ARRAY_SIZE(devconfs6);
+	char def_stable_secret[MAX_STR_CONF_LEN + 1] = {};
+	char all_stable_secret[MAX_STR_CONF_LEN + 1] = {};
 
-	netns.n_def_conf = ARRAY_SIZE(devconfs);
-	netns.n_all_conf = ARRAY_SIZE(devconfs);
-	netns.def_conf = xmalloc(sizeof(int) * netns.n_def_conf);
-	if (!netns.def_conf)
-		return -1;
-	netns.all_conf = xmalloc(sizeof(int) * netns.n_all_conf);
-	if (!netns.all_conf) {
-		xfree(netns.def_conf);
-		return -1;
+	netns.n_def_conf4 = size4;
+	netns.n_all_conf4 = size4;
+	netns.def_conf4 = xmalloc(sizeof(SysctlEntry *) * size4);
+	if (!netns.def_conf4)
+		goto err_free;
+	netns.all_conf4 = xmalloc(sizeof(SysctlEntry *) * size4);
+	if (!netns.all_conf4)
+		goto err_free;
+	def_confs4 = xmalloc(sizeof(SysctlEntry) * size4);
+	if (!def_confs4)
+		goto err_free;
+	all_confs4 = xmalloc(sizeof(SysctlEntry) * size4);
+	if (!all_confs4)
+		goto err_free;
+
+	for (i = 0; i < size4; i++) {
+		sysctl_entry__init(&def_confs4[i]);
+		sysctl_entry__init(&all_confs4[i]);
+		netns.def_conf4[i] = &def_confs4[i];
+		netns.all_conf4[i] = &all_confs4[i];
+		netns.def_conf4[i]->type = CTL_32;
+		netns.all_conf4[i]->type = CTL_32;
 	}
 
-	n = netns.n_def_conf;
-	ret = ipv4_conf_op("default", netns.def_conf, n, CTL_READ, NULL);
+	netns.n_def_conf6 = size6;
+	netns.n_all_conf6 = size6;
+	netns.def_conf6 = xmalloc(sizeof(SysctlEntry *) * size6);
+	if (!netns.def_conf6)
+		goto err_free;
+	netns.all_conf6 = xmalloc(sizeof(SysctlEntry *) * size6);
+	if (!netns.all_conf6)
+		goto err_free;
+	def_confs6 = xmalloc(sizeof(SysctlEntry) * size6);
+	if (!def_confs6)
+		goto err_free;
+	all_confs6 = xmalloc(sizeof(SysctlEntry) * size6);
+	if (!all_confs6)
+		goto err_free;
+
+	for (i = 0; i < size6; i++) {
+		sysctl_entry__init(&def_confs6[i]);
+		sysctl_entry__init(&all_confs6[i]);
+		netns.def_conf6[i] = &def_confs6[i];
+		netns.all_conf6[i] = &all_confs6[i];
+		if (strcmp(devconfs6[i], "stable_secret")) {
+			netns.def_conf6[i]->type = SYSCTL_TYPE__CTL_32;
+			netns.all_conf6[i]->type = SYSCTL_TYPE__CTL_32;
+		} else {
+			netns.def_conf6[i]->type = SYSCTL_TYPE__CTL_STR;
+			netns.all_conf6[i]->type = SYSCTL_TYPE__CTL_STR;
+			netns.def_conf6[i]->sarg = def_stable_secret;
+			netns.all_conf6[i]->sarg = all_stable_secret;
+		}
+	}
+
+	ret = ipv4_conf_op("default", netns.def_conf4, size4, CTL_READ, NULL);
 	if (ret < 0)
 		goto err_free;
-	ret = ipv4_conf_op("all", netns.all_conf, n, CTL_READ, NULL);
+	ret = ipv4_conf_op("all", netns.all_conf4, size4, CTL_READ, NULL);
+	if (ret < 0)
+		goto err_free;
+
+	ret = ipv6_conf_op("default", netns.def_conf6, size6, CTL_READ, NULL);
+	if (ret < 0)
+		goto err_free;
+	ret = ipv6_conf_op("all", netns.all_conf6, size6, CTL_READ, NULL);
 	if (ret < 0)
 		goto err_free;
 
 	ret = pb_write_one(img_from_set(fds, CR_FD_NETNS), &netns, PB_NETNS);
 err_free:
-	xfree(netns.def_conf);
-	xfree(netns.all_conf);
+	xfree(netns.def_conf4);
+	xfree(netns.all_conf4);
+	xfree(def_confs4);
+	xfree(all_confs4);
+	xfree(netns.def_conf6);
+	xfree(netns.all_conf6);
+	xfree(def_confs6);
+	xfree(all_confs6);
 	return ret;
 }
 
@@ -987,7 +1330,7 @@ out:
 
 static int restore_netns_conf(int pid, NetnsEntry **netns)
 {
-	int ret = 0, n;
+	int ret = 0;
 	struct cr_img *img;
 
 	img = open_image(CR_FD_NETNS, O_RSTR, pid);
@@ -1004,11 +1347,29 @@ static int restore_netns_conf(int pid, NetnsEntry **netns)
 		return -1;
 	}
 
-	n = (*netns)->n_def_conf;
-	ret = ipv4_conf_op("default", (*netns)->def_conf, n, CTL_WRITE, NULL);
-	if (ret)
-		goto out;
-	ret = ipv4_conf_op("all", (*netns)->all_conf, n, CTL_WRITE, NULL);
+	if ((*netns)->def_conf4) {
+		ret = ipv4_conf_op("all", (*netns)->all_conf4, (*netns)->n_all_conf4, CTL_WRITE, NULL);
+		if (ret)
+			goto out;
+		ret = ipv4_conf_op("default", (*netns)->def_conf4, (*netns)->n_def_conf4, CTL_WRITE, NULL);
+		if (ret)
+			goto out;
+	} else if ((*netns)->def_conf) {
+		/* Backward compatibility */
+		ret = ipv4_conf_op_old("all", (*netns)->all_conf, (*netns)->n_all_conf, CTL_WRITE, NULL);
+		if (ret)
+			goto out;
+		ret = ipv4_conf_op_old("default", (*netns)->def_conf, (*netns)->n_def_conf, CTL_WRITE, NULL);
+		if (ret)
+			goto out;
+	}
+
+	if ((*netns)->def_conf6) {
+		ret = ipv6_conf_op("all", (*netns)->all_conf6, (*netns)->n_all_conf6, CTL_WRITE, NULL);
+		if (ret)
+			goto out;
+		ret = ipv6_conf_op("default", (*netns)->def_conf6, (*netns)->n_def_conf6, CTL_WRITE, NULL);
+	}
 out:
 	close_image(img);
 	return ret;
@@ -1029,7 +1390,7 @@ static int mount_ns_sysfs(void)
 		return -1;
 	}
 
-	if (mount(NULL, "/", NULL, MS_PRIVATE | MS_REC, NULL)) {
+	if (mount(NULL, "/", NULL, MS_SLAVE | MS_REC, NULL)) {
 		pr_perror("Can't mark the root mount as private");
 		return -1;
 	}
@@ -1317,7 +1678,7 @@ static int prep_ns_sockets(struct ns_id *ns, bool for_dump)
 	} else
 		ns->net.nlsk = -1;
 
-	ret = ns->net.seqsk = socket(PF_UNIX, SOCK_SEQPACKET, 0);
+	ret = ns->net.seqsk = socket(PF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0);
 	if (ret < 0) {
 		pr_perror("Can't create seqsk for parasite");
 		goto err_sq;
