@@ -28,6 +28,8 @@
 #include "vma.h"
 #include "proc_parse.h"
 #include "aio.h"
+#include "fault-injection.h"
+#include "syscall-codes.h"
 
 #include <string.h>
 #include <stdlib.h>
@@ -44,7 +46,7 @@
 static int can_run_syscall(unsigned long ip, unsigned long start,
 			   unsigned long end, unsigned long pad)
 {
-	return ip >= start && ip < (end - code_syscall_size - pad);
+	return ip >= start && ip < (end - BUILTIN_SYSCALL_SIZE - pad);
 }
 
 static int syscall_fits_vma_area(struct vma_area *vma_area, unsigned long pad)
@@ -210,7 +212,8 @@ err:
 	return ret;
 }
 
-int __parasite_execute_syscall(struct parasite_ctl *ctl, user_regs_struct_t *regs)
+int __parasite_execute_syscall(struct parasite_ctl *ctl,
+		user_regs_struct_t *regs, const char *code_syscall)
 {
 	pid_t pid = ctl->pid.real;
 	int err;
@@ -400,7 +403,8 @@ static void sigchld_handler(int signal, siginfo_t *siginfo, void *data)
 	if (WIFEXITED(status))
 		pr_err("%d exited with %d unexpectedly\n", pid, WEXITSTATUS(status));
 	else if (WIFSIGNALED(status))
-		pr_err("%d was killed by %d unexpectedly\n", pid, WTERMSIG(status));
+		pr_err("%d was killed by %d unexpectedly: %s\n",
+			pid, WTERMSIG(status), strsignal(WTERMSIG(status)));
 	else if (WIFSTOPPED(status))
 		pr_err("%d was stopped by %d unexpectedly\n", pid, WSTOPSIG(status));
 
@@ -464,6 +468,10 @@ static int prepare_tsock(struct parasite_ctl *ctl, pid_t pid,
 		}
 	}
 
+	/* Check a case when parasite can't initialize a command socket */
+	if (fault_injected(FI_PARASITE_CONNECT))
+		args->h_addr_len = gen_parasite_saddr(&args->h_addr, getpid() + 1);
+
 	/*
 	 * Set to -1 to prevent any accidental misuse. The
 	 * only valid user of it is accept_tsock().
@@ -505,6 +513,8 @@ static int parasite_init_daemon(struct parasite_ctl *ctl, struct ns_id *net)
 	args->sigframe = ctl->rsigframe;
 	args->log_level = log_get_loglevel();
 
+	futex_set(&args->daemon_connected, 0);
+
 	if (prepare_tsock(ctl, pid, args, net))
 		goto err;
 
@@ -516,6 +526,13 @@ static int parasite_init_daemon(struct parasite_ctl *ctl, struct ns_id *net)
 	if (parasite_run(pid, PTRACE_CONT, ctl->parasite_ip, ctl->rstack, &regs, &ctl->orig))
 		goto err;
 
+	futex_wait_while_eq(&args->daemon_connected, 0);
+	if (futex_get(&args->daemon_connected) != 1) {
+		errno = -(int)futex_get(&args->daemon_connected);
+		pr_perror("Unable to connect a transport socket");
+		goto err;
+	}
+
 	if (accept_tsock(ctl) < 0)
 		goto err;
 
@@ -670,7 +687,8 @@ int parasite_dump_sigacts_seized(struct parasite_ctl *ctl, struct cr_imgset *cr_
 		ASSIGN_TYPED(se.sigaction, encode_pointer(args->sas[i].rt_sa_handler));
 		ASSIGN_TYPED(se.flags, args->sas[i].rt_sa_flags);
 		ASSIGN_TYPED(se.restorer, encode_pointer(args->sas[i].rt_sa_restorer));
-		ASSIGN_TYPED(se.mask, args->sas[i].rt_sa_mask.sig[0]);
+		BUILD_BUG_ON(sizeof(se.mask) != sizeof(args->sas[0].rt_sa_mask.sig));
+		memcpy(&se.mask, args->sas[i].rt_sa_mask.sig, sizeof(se.mask));
 
 		if (pb_write_one(img, &se, PB_SIGACT) < 0)
 			return -1;
@@ -808,14 +826,16 @@ struct parasite_tty_args *parasite_dump_tty(struct parasite_ctl *ctl, int fd, in
 }
 
 int parasite_drain_fds_seized(struct parasite_ctl *ctl,
-		struct parasite_drain_fd *dfds, int *lfds, struct fd_opts *opts)
+		struct parasite_drain_fd *dfds, int nr_fds, int off,
+		int *lfds, struct fd_opts *opts)
 {
 	int ret = -1, size;
 	struct parasite_drain_fd *args;
 
 	size = drain_fds_size(dfds);
 	args = parasite_args_s(ctl, size);
-	memcpy(args, dfds, size);
+	args->nr_fds = nr_fds;
+	memcpy(&args->fds, dfds->fds + off, sizeof(int) * nr_fds);
 
 	ret = __parasite_execute_daemon(PARASITE_CMD_DRAIN_FDS, ctl);
 	if (ret) {
@@ -823,7 +843,7 @@ int parasite_drain_fds_seized(struct parasite_ctl *ctl,
 		goto err;
 	}
 
-	ret = recv_fds(ctl->tsock, lfds, dfds->nr_fds, opts);
+	ret = recv_fds(ctl->tsock, lfds, nr_fds, opts);
 	if (ret)
 		pr_err("Can't retrieve FDs from socket\n");
 
@@ -927,6 +947,26 @@ static int parasite_fini_seized(struct parasite_ctl *ctl)
 	return 0;
 }
 
+static bool task_is_trapped(int status, pid_t pid)
+{
+	if (WIFSTOPPED(status) && WSTOPSIG(status) == SIGTRAP)
+		return true;
+
+	pr_err("Task %d is in unexpected state: %x\n", pid, status);
+	if (WIFEXITED(status))
+		pr_err("Task exited with %d\n", WEXITSTATUS(status));
+	if (WIFSIGNALED(status))
+		pr_err("Task signaled with %d: %s\n",
+			WTERMSIG(status), strsignal(WTERMSIG(status)));
+	if (WIFSTOPPED(status))
+		pr_err("Task stopped with %d: %s\n",
+			WSTOPSIG(status), strsignal(WSTOPSIG(status)));
+	if (WIFCONTINUED(status))
+		pr_err("Task continued\n");
+
+	return false;
+}
+
 /*
  * Trap tasks on the exit from the specified syscall
  *
@@ -950,10 +990,8 @@ int parasite_stop_on_syscall(int tasks, const int sys_nr, enum trace_flags trace
 			return -1;
 		}
 
-		if (!WIFSTOPPED(status) || WSTOPSIG(status) != SIGTRAP) {
-			pr_err("Task is in unexpected state: %x\n", status);
+		if (!task_is_trapped(status, pid))
 			return -1;
-		}
 
 		pr_debug("%d was trapped\n", pid);
 
@@ -971,7 +1009,7 @@ int parasite_stop_on_syscall(int tasks, const int sys_nr, enum trace_flags trace
 			return -1;
 		}
 
-		pr_debug("%d is going to execute the syscall %lx\n", pid, REG_SYSCALL_NR(regs));
+		pr_debug("%d is going to execute the syscall %lu\n", pid, REG_SYSCALL_NR(regs));
 		if (REG_SYSCALL_NR(regs) == sys_nr) {
 			/*
 			 * The process is going to execute the required syscall,
@@ -989,10 +1027,8 @@ int parasite_stop_on_syscall(int tasks, const int sys_nr, enum trace_flags trace
 				return -1;
 			}
 
-			if (!WIFSTOPPED(status) || WSTOPSIG(status) != SIGTRAP) {
-				pr_err("Task is in unexpected state: %x\n", status);
+			if (!task_is_trapped(status, pid))
 				return -1;
-			}
 
 			pr_debug("%d was stopped\n", pid);
 			tasks--;
@@ -1032,12 +1068,14 @@ int parasite_stop_daemon(struct parasite_ctl *ctl)
 
 int parasite_cure_remote(struct parasite_ctl *ctl)
 {
-	int ret = 0;
-
 	if (parasite_stop_daemon(ctl))
 		return -1;
 
-	if (ctl->remote_map) {
+	if (!ctl->remote_map)
+		return 0;
+
+	/* Unseizing task with parasite -- it does it himself */
+	if (ctl->addr_cmd) {
 		struct parasite_unmap_args *args;
 
 		*ctl->addr_cmd = PARASITE_CMD_UNMAP;
@@ -1046,10 +1084,21 @@ int parasite_cure_remote(struct parasite_ctl *ctl)
 		args->parasite_start = ctl->remote_map;
 		args->parasite_len = ctl->map_length;
 		if (parasite_unmap(ctl, ctl->parasite_ip))
-			ret = -1;
+			return -1;
+	} else {
+		unsigned long ret;
+
+		syscall_seized(ctl, __NR_munmap, &ret,
+				(unsigned long)ctl->remote_map, ctl->map_length,
+				0, 0, 0, 0);
+		if (ret) {
+			pr_err("munmap for remote map %p, %lu returned %lu\n",
+					ctl->remote_map, ctl->map_length, ret);
+			return -1;
+		}
 	}
 
-	return ret;
+	return 0;
 }
 
 int parasite_cure_local(struct parasite_ctl *ctl)
@@ -1089,7 +1138,7 @@ int parasite_unmap(struct parasite_ctl *ctl, unsigned long addr)
 	pid_t pid = ctl->pid.real;
 	int ret = -1;
 
-	ret = parasite_run(pid, PTRACE_SYSCALL, addr, NULL, &regs, &ctl->orig);
+	ret = parasite_run(pid, PTRACE_SYSCALL, addr, ctl->rstack, &regs, &ctl->orig);
 	if (ret)
 		goto err;
 
@@ -1181,7 +1230,6 @@ static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size)
 	return 0;
 }
 
-#ifdef CONFIG_HAS_MEMFD
 static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 {
 	void *where = (void *)ctl->syscall_ip + BUILTIN_SYSCALL_SIZE;
@@ -1190,6 +1238,9 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 	unsigned long sret = -ENOSYS;
 	int ret, fd, lfd;
 
+	if (fault_injected(FI_NO_MEMFD))
+		return 1;
+
 	BUILD_BUG_ON(sizeof(orig_code) < sizeof(long));
 
 	if (ptrace_swap_area(pid, where, (void *)orig_code, sizeof(orig_code))) {
@@ -1255,12 +1306,6 @@ err_cure:
 	syscall_seized(ctl, __NR_close, &sret, fd, 0, 0, 0, 0, 0);
 	return -1;
 }
-#else
-static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
-{
-	return 1;
-}
-#endif
 
 int parasite_map_exchange(struct parasite_ctl *ctl, unsigned long size)
 {
@@ -1384,6 +1429,8 @@ struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 	if (parasite_start_daemon(ctl, item))
 		goto err_restore;
 
+	dmpi(item)->parasite_ctl = ctl;
+
 	return ctl;
 
 err_restore:
@@ -1395,7 +1442,11 @@ int ptrace_stop_pie(pid_t pid, void *addr, enum trace_flags *tf)
 {
 	int ret;
 
-	ret = ptrace_set_breakpoint(pid, addr);
+	if (fault_injected(FI_NO_BREAKPOINTS)) {
+		pr_debug("Force no-breakpoints restore\n");
+		ret = 0;
+	} else
+		ret = ptrace_set_breakpoint(pid, addr);
 	if (ret < 0)
 		return ret;
 
