@@ -3,8 +3,10 @@
 
 #include <linux/securebits.h>
 #include <linux/capability.h>
+#include <linux/aio_abi.h>
 #include <sys/types.h>
 #include <sys/mman.h>
+#include <sys/file.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/time.h>
@@ -16,6 +18,7 @@
 #include <signal.h>
 
 #include "compiler.h"
+#include "asm/string.h"
 #include "asm/types.h"
 #include "syscall.h"
 #include "config.h"
@@ -26,7 +29,6 @@
 #include "sk-inet.h"
 #include "vma.h"
 
-#include "crtools.h"
 #include "lock.h"
 #include "restorer.h"
 #include "aio.h"
@@ -35,6 +37,7 @@
 #include "images/creds.pb-c.h"
 #include "images/mm.pb-c.h"
 
+#include "shmem.h"
 #include "asm/restorer.h"
 
 #ifndef PR_SET_PDEATHSIG
@@ -49,7 +52,7 @@
 		__ret;							\
 	})
 
-static struct task_entries *task_entries;
+static struct task_entries *task_entries_local;
 static futex_t thread_inprogress;
 static pid_t *helpers;
 static int n_helpers;
@@ -83,7 +86,7 @@ static void sigchld_handler(int signal, siginfo_t *siginfo, void *data)
 
 	pr_info("Task %d %s %d\n", siginfo->si_pid, r, siginfo->si_status);
 
-	futex_abort_and_wake(&task_entries->nr_in_progress);
+	futex_abort_and_wake(&task_entries_local->nr_in_progress);
 	/* sa_restorer may be unmaped, so we can't go back to userspace*/
 	sys_kill(sys_getpid(), SIGSTOP);
 	sys_exit_group(1);
@@ -394,8 +397,7 @@ die:
 	return -1;
 }
 
-static int restore_thread_common(struct rt_sigframe *sigframe,
-		struct thread_restore_args *args)
+static int restore_thread_common(struct thread_restore_args *args)
 {
 	sys_set_tid_address((int *)decode_pointer(args->clear_tid_addr));
 
@@ -450,9 +452,9 @@ long __export_restore_thread(struct thread_restore_args *args)
 		goto core_restore_end;
 	}
 
-	rt_sigframe = (void *)args->mem_zone.rt_sigframe;
+	rt_sigframe = (void *)&args->mz->rt_sigframe;
 
-	if (restore_thread_common(rt_sigframe, args))
+	if (restore_thread_common(args))
 		goto core_restore_end;
 
 	ret = restore_creds(args->creds_args, args->ta->proc_fd);
@@ -465,27 +467,27 @@ long __export_restore_thread(struct thread_restore_args *args)
 
 	pr_info("%ld: Restored\n", sys_gettid());
 
-	restore_finish_stage(CR_STATE_RESTORE);
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE);
 
 	if (restore_signals(args->siginfo, args->siginfo_n, false))
 		goto core_restore_end;
 
-	restore_finish_stage(CR_STATE_RESTORE_SIGCHLD);
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE_SIGCHLD);
 	restore_pdeath_sig(args);
 
 	if (args->ta->seccomp_mode != SECCOMP_MODE_DISABLED)
 		pr_info("Restoring seccomp mode %d for %ld\n", args->ta->seccomp_mode, sys_getpid());
 
-	restore_finish_stage(CR_STATE_RESTORE_CREDS);
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE_CREDS);
 
 	futex_dec_and_wake(&thread_inprogress);
 
-	new_sp = (long)rt_sigframe + SIGFRAME_OFFSET;
+	new_sp = (long)rt_sigframe + RT_SIGFRAME_OFFSET(rt_sigframe);
 	rst_sigreturn(new_sp);
 
 core_restore_end:
 	pr_err("Restorer abnormal termination for %ld\n", sys_getpid());
-	futex_abort_and_wake(&task_entries->nr_in_progress);
+	futex_abort_and_wake(&task_entries_local->nr_in_progress);
 	sys_exit_group(1);
 	return -1;
 }
@@ -503,15 +505,30 @@ static long restore_self_exe_late(struct task_restore_args *args)
 	return ret;
 }
 
-static unsigned long restore_mapping(const VmaEntry *vma_entry)
+static unsigned long restore_mapping(VmaEntry *vma_entry)
 {
 	int prot	= vma_entry->prot;
 	int flags	= vma_entry->flags | MAP_FIXED;
 	unsigned long addr;
 
-	if (vma_entry_is(vma_entry, VMA_AREA_SYSVIPC))
-		return sys_shmat(vma_entry->fd, decode_pointer(vma_entry->start),
-				 (vma_entry->prot & PROT_WRITE) ? 0 : SHM_RDONLY);
+	if (vma_entry_is(vma_entry, VMA_AREA_SYSVIPC)) {
+		int att_flags;
+		/*
+		 * See comment in open_shmem_sysv() for what SYSV_SHMEM_SKIP_FD
+		 * means and why we check for PROT_EXEC few lines below.
+		 */
+		if (vma_entry->fd == SYSV_SHMEM_SKIP_FD)
+			return vma_entry->start;
+
+		if (vma_entry->prot & PROT_EXEC) {
+			att_flags = 0;
+			vma_entry->prot &= ~PROT_EXEC;
+		} else
+			att_flags = SHM_RDONLY;
+
+		pr_info("Attach SYSV shmem %d at %"PRIx64"\n", (int)vma_entry->fd, vma_entry->start);
+		return sys_shmat(vma_entry->fd, decode_pointer(vma_entry->start), att_flags);
+	}
 
 	/*
 	 * Restore or shared mappings are tricky, since
@@ -546,6 +563,137 @@ static unsigned long restore_mapping(const VmaEntry *vma_entry)
 	return addr;
 }
 
+/*
+ * This restores aio ring header, content, head and in-kernel position
+ * of tail. To set tail, we write to /dev/null and use the fact this
+ * operation is synchronious for the device. Also, we unmap temporary
+ * anonymous area, used to store content of ring buffer during restore
+ * and mapped in map_private_vma().
+ */
+static int restore_aio_ring(struct rst_aio_ring *raio)
+{
+	struct aio_ring *ring = (void *)raio->addr, *new;
+	int i, maxr, count, fd, ret;
+	unsigned head = ring->head;
+	unsigned tail = ring->tail;
+	struct iocb *iocb, **iocbp;
+	unsigned long ctx = 0;
+	unsigned size;
+	char buf[1];
+
+	ret = sys_io_setup(raio->nr_req, &ctx);
+	if (ret < 0) {
+		pr_err("Ring setup failed with %d\n", ret);
+		return -1;
+	}
+
+	new = (struct aio_ring *)ctx;
+	i = (raio->len - sizeof(struct aio_ring)) / sizeof(struct io_event);
+	if (tail >= ring->nr || head >= ring->nr || ring->nr != i ||
+	    new->nr != ring->nr) {
+		pr_err("wrong aio: tail=%x head=%x req=%x old_nr=%x new_nr=%x expect=%x\n",
+			tail, head, raio->nr_req, ring->nr, new->nr, i);
+
+		return -1;
+	}
+
+	if (tail == 0 && head == 0)
+		goto populate;
+
+	fd = sys_open("/dev/null", O_WRONLY, 0);
+	if (fd < 0) {
+		pr_err("Can't open /dev/null for aio\n");
+		return -1;
+	}
+
+	/*
+	 * If tail < head, we have to do full turn and then submit
+	 * tail more request, i.e. ring->nr + tail.
+	 * If we do not do full turn, in-kernel completed_events
+	 * will initialize wrong.
+	 *
+	 * Maximum number reqs to submit at once are ring->nr-1,
+	 * so we won't allocate more.
+	 */
+	if (tail < head)
+		count = ring->nr + tail;
+	else
+		count = tail;
+	maxr = min_t(unsigned, count, ring->nr-1);
+
+	/*
+	 * Since we only interested in moving the tail, the requests
+	 * may be any. We submit count identical requests.
+	 */
+	size = sizeof(struct iocb) + maxr * sizeof(struct iocb *);
+	iocb = (void *)sys_mmap(NULL, size, PROT_READ|PROT_WRITE,
+				MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+	iocbp = (void *)iocb + sizeof(struct iocb);
+
+	if (IS_ERR(iocb)) {
+		pr_err("Can't mmap aio tmp buffer: %ld\n", PTR_ERR(iocb));
+		return -1;
+	}
+
+	iocb->aio_fildes = fd;
+	iocb->aio_buf = (unsigned long)buf;
+	iocb->aio_nbytes = 1;
+	iocb->aio_lio_opcode = IOCB_CMD_PWRITE; /* Write is nop, read populates buf */
+
+	for (i = 0; i < maxr; i++)
+		iocbp[i] = iocb;
+
+	i = 0;
+	do {
+		ret = sys_io_submit(ctx, count - i, iocbp);
+		if (ret < 0) {
+			pr_err("Can't submit aio iocbs: ret=%d\n", ret);
+			return -1;
+		}
+		i += ret;
+
+		 /*
+		  * We may submit less than requested, because of too big
+		  * count OR behaviour of get_reqs_available(), which
+		  * takes available requests only if their number is
+		  * aliquot to kioctx::req_batch. Free part of buffer
+		  * for next iteration.
+		  *
+		  * Direct set of head is equal to sys_io_getevents() call,
+		  * and faster. See kernel for the details.
+		  */
+		((struct aio_ring *)ctx)->head = i < head ? i : head;
+	} while (i < count);
+
+	sys_munmap(iocb, size);
+	sys_close(fd);
+
+populate:
+	i = offsetof(struct aio_ring, io_events);
+	builtin_memcpy((void *)ctx + i, (void *)ring + i, raio->len - i);
+
+	/*
+	 * If we failed to get the proper nr_req right and
+	 * created smaller or larger ring, then this remap
+	 * will (should) fail, since AIO rings has immutable
+	 * size.
+	 *
+	 * This is not great, but anyway better than putting
+	 * a ring of wrong size into correct place.
+	 *
+	 * Also, this unmaps temporary anonymous area on raio->addr.
+	 */
+
+	ctx = sys_mremap(ctx, raio->len, raio->len,
+				MREMAP_FIXED | MREMAP_MAYMOVE,
+				raio->addr);
+	if (ctx != raio->addr) {
+		pr_err("Ring remap failed with %ld\n", ctx);
+		return -1;
+	}
+	return 0;
+}
+
 static void rst_tcp_repair_off(struct rst_tcp_sock *rts)
 {
 	int aux, ret;
@@ -738,7 +886,7 @@ static void *bootstrap_start;
 static unsigned int bootstrap_len;
 
 /*
- * sys_munmap must not return here. The controll process must
+ * sys_munmap must not return here. The control process must
  * trap us on the exit from sys_munmap.
  */
 #ifdef CONFIG_VDSO
@@ -837,16 +985,28 @@ static int wait_zombies(struct task_restore_args *task_args)
 {
 	int i;
 
-	atomic_add(task_args->zombies_n, &task_entries->nr_zombies);
-
 	for (i = 0; i < task_args->zombies_n; i++) {
-		if (sys_waitid(P_PID, task_args->zombies[i], NULL, WNOWAIT | WEXITED, NULL) < 0) {
-			pr_err("Wait on %d zombie failed\n", task_args->zombies[i]);
+		int ret, nr_in_progress;
+
+		nr_in_progress = futex_get(&task_entries_local->nr_in_progress);
+
+		ret = sys_waitid(P_PID, task_args->zombies[i], NULL, WNOWAIT | WEXITED, NULL);
+		if (ret == -ECHILD) {
+			/* A process isn't reparented to this task yet.
+			 * Let's wait when someone complete this stage
+			 * and try again.
+			 */
+			futex_wait_while_eq(&task_entries_local->nr_in_progress,
+								nr_in_progress);
+			i--;
+			continue;
+		}
+		if (ret < 0) {
+			pr_err("Wait on %d zombie failed: %d\n", task_args->zombies[i], ret);
 			return -1;
 		}
 		pr_debug("%ld: Collect a zombie with pid %d\n",
 			sys_getpid(), task_args->zombies[i]);
-		futex_dec_and_wake(&task_entries->nr_in_progress);
 	}
 
 	return 0;
@@ -880,7 +1040,7 @@ long __export_restore_task(struct task_restore_args *args)
 	vdso_rt_size	= args->vdso_rt_size;
 #endif
 
-	task_entries = args->task_entries;
+	task_entries_local = args->task_entries;
 	helpers = args->helpers;
 	n_helpers = args->helpers_n;
 	zombies = args->zombies;
@@ -1000,54 +1160,9 @@ long __export_restore_task(struct task_restore_args *args)
 	 * up AIO rings.
 	 */
 
-	for (i = 0; i < args->rings_n; i++) {
-		struct rst_aio_ring *raio = &args->rings[i];
-		unsigned long ctx = 0;
-		int ret;
-
-		ret = sys_io_setup(raio->nr_req, &ctx);
-		if (ret < 0) {
-			pr_err("Ring setup failed with %d\n", ret);
+	for (i = 0; i < args->rings_n; i++)
+		if (restore_aio_ring(&args->rings[i]) < 0)
 			goto core_restore_end;
-		}
-
-		if (ctx == raio->addr) /* Lucky bastards we are! */
-			continue;
-
-		/*
-		 * If we failed to get the proper nr_req right and
-		 * created smaller or larger ring, then this remap
-		 * will (should) fail, since AIO rings has immutable
-		 * size.
-		 *
-		 * This is not great, but anyway better than putting
-		 * a ring of wrong size into correct place.
-		 */
-
-		ctx = sys_mremap(ctx, raio->len, raio->len,
-					MREMAP_FIXED | MREMAP_MAYMOVE,
-					raio->addr);
-		if (ctx != raio->addr) {
-			pr_err("Ring remap failed with %ld\n", ctx);
-			goto core_restore_end;
-		}
-
-		/*
-		 * Now check that kernel not just remapped the
-		 * ring into new place, but updated the internal
-		 * context state respectively.
-		 */
-
-		ret = sys_io_getevents(ctx, 0, 1, NULL, NULL);
-		if (ret != 0) {
-			if (ret < 0)
-				pr_err("Kernel doesn't remap AIO rings\n");
-			else
-				pr_err("AIO context screwed up\n");
-
-			goto core_restore_end;
-		}
-	}
 
 	/*
 	 * Finally restore madivse() bits
@@ -1140,9 +1255,9 @@ long __export_restore_task(struct task_restore_args *args)
 	 * registers from the frame, set them up and
 	 * finally pass execution to the new IP.
 	 */
-	rt_sigframe = (void *)args->t->mem_zone.rt_sigframe;
+	rt_sigframe = (void *)&args->t->mz->rt_sigframe;
 
-	if (restore_thread_common(rt_sigframe, args->t))
+	if (restore_thread_common(args->t))
 		goto core_restore_end;
 
 	/*
@@ -1191,7 +1306,7 @@ long __export_restore_task(struct task_restore_args *args)
 			if (thread_args[i].pid == args->t->pid)
 				continue;
 
-			new_sp = restorer_stack(thread_args + i);
+			new_sp = restorer_stack(thread_args[i].mz);
 			last_pid_len = vprint_num(last_pid_buf, sizeof(last_pid_buf), thread_args[i].pid - 1, &s);
 			sys_lseek(fd, 0, SEEK_SET);
 			ret = sys_write(fd, s, last_pid_len);
@@ -1237,13 +1352,12 @@ long __export_restore_task(struct task_restore_args *args)
 
 	pr_info("%ld: Restored\n", sys_getpid());
 
-	restore_finish_stage(CR_STATE_RESTORE);
-
-	if (wait_zombies(args) < 0)
-		goto core_restore_end;
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE);
 
 	if (wait_helpers(args) < 0)
 		goto core_restore_end;
+	if (wait_zombies(args) < 0)
+		goto core_restore_end;
 
 	ksigfillset(&to_block);
 	ret = sys_sigprocmask(SIG_SETMASK, &to_block, NULL, sizeof(k_rtsigset_t));
@@ -1262,7 +1376,7 @@ long __export_restore_task(struct task_restore_args *args)
 	if (ret)
 		goto core_restore_end;
 
-	restore_finish_stage(CR_STATE_RESTORE_SIGCHLD);
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE_SIGCHLD);
 
 	rst_tcp_socks_all(args);
 
@@ -1284,7 +1398,7 @@ long __export_restore_task(struct task_restore_args *args)
 
 	futex_set_and_wake(&thread_inprogress, args->nr_threads);
 
-	restore_finish_stage(CR_STATE_RESTORE_CREDS);
+	restore_finish_stage(task_entries_local, CR_STATE_RESTORE_CREDS);
 
 	if (ret)
 		BUG();
@@ -1318,7 +1432,7 @@ long __export_restore_task(struct task_restore_args *args)
 	/*
 	 * Sigframe stack.
 	 */
-	new_sp = (long)rt_sigframe + SIGFRAME_OFFSET;
+	new_sp = (long)rt_sigframe + RT_SIGFRAME_OFFSET(rt_sigframe);
 
 	/*
 	 * Prepare the stack and call for sigreturn,
@@ -1328,7 +1442,7 @@ long __export_restore_task(struct task_restore_args *args)
 	rst_sigreturn(new_sp);
 
 core_restore_end:
-	futex_abort_and_wake(&task_entries->nr_in_progress);
+	futex_abort_and_wake(&task_entries_local->nr_in_progress);
 	pr_err("Restorer fail %ld\n", sys_getpid());
 	sys_exit_group(1);
 	return -1;
