@@ -5,7 +5,7 @@
 
 #include "cr_options.h"
 #include "pstree.h"
-#include "util.h"
+#include "rst-malloc.h"
 #include "lock.h"
 #include "namespaces.h"
 #include "files.h"
@@ -17,8 +17,7 @@
 #include "images/pstree.pb-c.h"
 
 struct pstree_item *root_item;
-
-#define CLONE_ALLNS     (CLONE_NEWPID | CLONE_NEWNET | CLONE_NEWIPC | CLONE_NEWUTS | CLONE_NEWNS | CLONE_NEWUSER)
+static struct rb_root pid_root_rb;
 
 void core_entry_free(CoreEntry *core)
 {
@@ -164,7 +163,8 @@ void pstree_free_cores(struct pstree_item *item)
 
 	if (item->core) {
 		for (i = 1; i < item->nr_threads; i++)
-			core_entry_free(item->core[i]);
+			if (item->core[i])
+				core_entry_free(item->core[i]);
 		xfree(item->core);
 		item->core = NULL;
 	}
@@ -219,18 +219,11 @@ struct pstree_item *__alloc_pstree_item(bool rst)
 	return item;
 }
 
-struct pstree_item *alloc_pstree_helper(void)
+void init_pstree_helper(struct pstree_item *ret)
 {
-	struct pstree_item *ret;
-
-	ret = alloc_pstree_item_with_rst();
-	if (ret) {
-		ret->state = TASK_HELPER;
-		rsti(ret)->clone_flags = CLONE_FILES | CLONE_FS;
-		task_entries->nr_helpers++;
-	}
-
-	return ret;
+	ret->pid.state = TASK_HELPER;
+	rsti(ret)->clone_flags = CLONE_FILES | CLONE_FS;
+	task_entries->nr_helpers++;
 }
 
 /* Deep first search on children */
@@ -326,8 +319,6 @@ err:
 	return ret;
 }
 
-static int max_pid = 0;
-
 static int prepare_pstree_for_shell_job(void)
 {
 	pid_t current_sid = getsid(getpid());
@@ -374,17 +365,117 @@ static int prepare_pstree_for_shell_job(void)
 			pi->sid = current_sid;
 	}
 
-	max_pid = max((int)current_sid, max_pid);
-	max_pid = max((int)current_gid, max_pid);
+	if (lookup_create_item(current_sid) == NULL)
+		return -1;
+	if (lookup_create_item(current_gid) == NULL)
+		return -1;
 
 	return 0;
 }
 
-static int read_pstree_image(void)
+/*
+ * Try to find a pid node in the tree and insert a new one,
+ * it is not there yet. If pid_node isn't set, pstree_item
+ * is inserted.
+ */
+static struct pid *lookup_create_pid(pid_t pid, struct pid *pid_node)
+{
+	struct rb_node *node = pid_root_rb.rb_node;
+	struct rb_node **new = &pid_root_rb.rb_node;
+	struct rb_node *parent = NULL;
+
+	while (node) {
+		struct pid *this = rb_entry(node, struct pid, node);
+
+		parent = *new;
+		if (pid < this->virt)
+			node = node->rb_left, new = &((*new)->rb_left);
+		else if (pid > this->virt)
+			node = node->rb_right, new = &((*new)->rb_right);
+		else
+			return this;
+	}
+
+	if (!pid_node) {
+		struct pstree_item *item;
+
+		item = __alloc_pstree_item(true);
+		if (item == NULL)
+			return NULL;
+
+		item->pid.virt = pid;
+		pid_node = &item->pid;
+	}
+	rb_link_and_balance(&pid_root_rb, &pid_node->node, parent, new);
+	return pid_node;
+}
+
+void pstree_insert_pid(pid_t pid, struct pid *pid_node)
+{
+	struct pid* n;
+
+	n = lookup_create_pid(pid, pid_node);
+
+	BUG_ON(n != pid_node);
+}
+
+struct pstree_item *lookup_create_item(pid_t pid)
+{
+	struct pid *node;;
+
+	node = lookup_create_pid(pid, NULL);
+	if (!node)
+		return NULL;
+	BUG_ON(node->state == TASK_THREAD);
+
+	return container_of(node, struct pstree_item, pid);
+}
+
+struct pid *pstree_pid_by_virt(pid_t pid)
+{
+	struct rb_node *node = pid_root_rb.rb_node;
+
+	while (node) {
+		struct pid *this = rb_entry(node, struct pid, node);
+
+		if (pid < this->virt)
+			node = node->rb_left;
+		else if (pid > this->virt)
+			node = node->rb_right;
+		else
+			return this;
+	}
+	return NULL;
+}
+
+static int read_pstree_ids(struct pstree_item *pi)
+{
+	int ret;
+	struct cr_img *img;
+
+	img = open_image(CR_FD_IDS, O_RSTR, pi->pid.virt);
+	if (!img)
+		return -1;
+
+	ret = pb_read_one_eof(img, &pi->ids, PB_IDS);
+	close_image(img);
+
+	if (ret <= 0)
+		return ret;
+
+	if (pi->ids->has_mnt_ns_id) {
+		if (rst_add_ns_id(pi->ids->mnt_ns_id, pi, &mnt_ns_desc))
+			return -1;
+	}
+
+	return 0;
+}
+
+static int read_pstree_image(pid_t *pid_max)
 {
 	int ret = 0, i;
 	struct cr_img *img;
-	struct pstree_item *pi, *parent = NULL;
+	struct pstree_item *pi;
 
 	pr_info("Reading image tree\n");
 
@@ -400,18 +491,32 @@ static int read_pstree_image(void)
 			break;
 
 		ret = -1;
-		pi = alloc_pstree_item_with_rst();
+		pi = lookup_create_item(e->pid);
 		if (pi == NULL)
 			break;
+		BUG_ON(pi->pid.state != TASK_UNDEF);
+
+		/*
+		 * All pids should be added in the tree to be able to find
+		 * free pid-s for helpers. pstree_item for these pid-s will
+		 * be initialized when we meet PstreeEntry with this pid or
+		 * we will create helpers for them.
+		 */
+		if (lookup_create_item(e->pgid) == NULL)
+			break;
+		if (lookup_create_item(e->sid) == NULL)
+			break;
 
 		pi->pid.virt = e->pid;
-		max_pid = max((int)e->pid, max_pid);
-
+		if (e->pid > *pid_max)
+			*pid_max = e->pid;
 		pi->pgid = e->pgid;
-		max_pid = max((int)e->pgid, max_pid);
-
+		if (e->pgid > *pid_max)
+			*pid_max = e->pgid;
 		pi->sid = e->sid;
-		max_pid = max((int)e->sid, max_pid);
+		if (e->sid > *pid_max)
+			*pid_max = e->sid;
+		pi->pid.state = TASK_ALIVE;
 
 		if (e->ppid == 0) {
 			if (root_item) {
@@ -422,46 +527,41 @@ static int read_pstree_image(void)
 			root_item = pi;
 			pi->parent = NULL;
 		} else {
-			/*
-			 * Fast path -- if the pstree image is not edited, the
-			 * parent of any item should have already being restored
-			 * and sit among the last item's ancestors.
-			 */
-			while (parent) {
-				if (parent->pid.virt == e->ppid)
-					break;
-				parent = parent->parent;
-			}
+			struct pid *pid;
+			struct pstree_item *parent;
 
-			if (parent == NULL) {
-				for_each_pstree_item(parent) {
-					if (parent->pid.virt == e->ppid)
-						break;
-				}
-
-				if (parent == NULL) {
-					pr_err("Can't find a parent for %d\n", pi->pid.virt);
-					pstree_entry__free_unpacked(e, NULL);
-					xfree(pi);
-					goto err;
-				}
+			pid = pstree_pid_by_virt(e->ppid);
+			if (!pid || pid->state == TASK_UNDEF || pid->state == TASK_THREAD) {
+				pr_err("Can't find a parent for %d\n", pi->pid.virt);
+				pstree_entry__free_unpacked(e, NULL);
+				xfree(pi);
+				goto err;
 			}
 
+			parent = container_of(pid, struct pstree_item, pid);
 			pi->parent = parent;
 			list_add(&pi->sibling, &parent->children);
 		}
 
-		parent = pi;
-
 		pi->nr_threads = e->n_threads;
 		pi->threads = xmalloc(e->n_threads * sizeof(struct pid));
 		if (!pi->threads)
 			break;
 
 		for (i = 0; i < e->n_threads; i++) {
+			struct pid *node;
 			pi->threads[i].real = -1;
 			pi->threads[i].virt = e->threads[i];
-			max_pid = max((int)e->threads[i], max_pid);
+			pi->threads[i].state = TASK_THREAD;
+			if (i == 0)
+				continue; /* A thread leader is in a tree already */
+			node = lookup_create_pid(pi->threads[i].virt, &pi->threads[i]);
+
+			BUG_ON(node == NULL);
+			if (node != &pi->threads[i]) {
+				pr_err("Unexpected task %d in a tree %d\n", e->threads[i], i);
+				return -1;
+			}
 		}
 
 		task_entries->nr_threads += e->n_threads;
@@ -469,31 +569,43 @@ static int read_pstree_image(void)
 
 		pstree_entry__free_unpacked(e, NULL);
 
-		{
-			struct cr_img *img;
-
-			img = open_image(CR_FD_IDS, O_RSTR, pi->pid.virt);
-			if (!img)
-				goto err;
-			ret = pb_read_one_eof(img, &pi->ids, PB_IDS);
-			close_image(img);
-		}
-
-		if (ret == 0)
-			continue;
+		ret = read_pstree_ids(pi);
 		if (ret < 0)
 			goto err;
-
-		if (pi->ids->has_mnt_ns_id) {
-			if (rst_add_ns_id(pi->ids->mnt_ns_id, pi, &mnt_ns_desc))
-				goto err;
-		}
 	}
+
 err:
 	close_image(img);
 	return ret;
 }
 
+#define RESERVED_PIDS           300
+static int get_free_pid()
+{
+	static struct pstree_item *prev, *next;
+
+	if (prev == NULL)
+		prev = rb_entry(rb_first(&pid_root_rb), struct pstree_item, pid.node);
+
+	while (1) {
+		struct rb_node *node;
+		pid_t pid;
+
+		pid = prev->pid.virt + 1;
+		pid = pid < RESERVED_PIDS ? RESERVED_PIDS + 1 : pid;
+
+		node = rb_next(&prev->pid.node);
+		if (node == NULL)
+			return pid;
+		next = rb_entry(node, struct pstree_item, pid.node);
+		if (next->pid.virt > pid)
+			return pid;
+		prev = next;
+	}
+
+	return -1;
+}
+
 static int prepare_pstree_ids(void)
 {
 	struct pstree_item *item, *child, *helper, *tmp;
@@ -508,6 +620,7 @@ static int prepare_pstree_ids(void)
 	 * reparented to init.
 	 */
 	list_for_each_entry(item, &root_item->children, sibling) {
+		struct pstree_item *leader;
 
 		/*
 		 * If a child belongs to the root task's session or it's
@@ -517,15 +630,37 @@ static int prepare_pstree_ids(void)
 		if (item->sid == root_item->sid || item->sid == item->pid.virt)
 			continue;
 
-		helper = alloc_pstree_helper();
-		if (helper == NULL)
-			return -1;
-		helper->sid = item->sid;
-		helper->pgid = item->sid;
-		helper->pid.virt = item->sid;
-		helper->parent = root_item;
-		helper->ids = root_item->ids;
-		list_add_tail(&helper->sibling, &helpers);
+		leader = pstree_item_by_virt(item->sid);
+		BUG_ON(leader == NULL);
+		if (leader->pid.state != TASK_UNDEF) {
+			pid_t pid;
+
+			pid = get_free_pid();
+			if (pid < 0)
+				break;
+			helper = lookup_create_item(pid);
+			if (helper == NULL)
+				return -1;
+
+			pr_info("Session leader %d\n", item->sid);
+
+			helper->sid = item->sid;
+			helper->pgid = leader->pgid;
+			helper->ids = leader->ids;
+			helper->parent = leader;
+			list_add(&helper->sibling, &leader->children);
+
+			pr_info("Attach %d to the task %d\n",
+					helper->pid.virt, leader->pid.virt);
+		} else {
+			helper = leader;
+			helper->sid = item->sid;
+			helper->pgid = item->sid;
+			helper->parent = root_item;
+			helper->ids = root_item->ids;
+			list_add_tail(&helper->sibling, &helpers);
+		}
+		init_pstree_helper(helper);
 
 		pr_info("Add a helper %d for restoring SID %d\n",
 				helper->pid.virt, helper->sid);
@@ -555,7 +690,7 @@ static int prepare_pstree_ids(void)
 		if (!item->parent) /* skip the root task */
 			continue;
 
-		if (item->state == TASK_HELPER)
+		if (item->pid.state == TASK_HELPER)
 			continue;
 
 		if (item->sid != item->pid.virt) {
@@ -585,27 +720,6 @@ static int prepare_pstree_ids(void)
 
 			continue;
 		}
-
-		pr_info("Session leader %d\n", item->sid);
-
-		/* Try to find helpers, who should be connected to the leader */
-		list_for_each_entry(child, &helpers, sibling) {
-			if (child->state != TASK_HELPER)
-				continue;
-
-			if (child->sid != item->sid)
-				continue;
-
-			child->pgid = item->pgid;
-			child->pid.virt = ++max_pid;
-			child->parent = item;
-			list_move(&child->sibling, &item->children);
-
-			pr_info("Attach %d to the task %d\n",
-					child->pid.virt, item->pid.virt);
-
-			break;
-		}
 	}
 
 	/* All other helpers are session leaders for own sessions */
@@ -613,18 +727,15 @@ static int prepare_pstree_ids(void)
 
 	/* Add a process group leader if it is absent  */
 	for_each_pstree_item(item) {
-		struct pstree_item *gleader;
+		struct pid *pid;
 
 		if (!item->pgid || item->pid.virt == item->pgid)
 			continue;
 
-		for_each_pstree_item(gleader) {
-			if (gleader->pid.virt == item->pgid)
-				break;
-		}
-
-		if (gleader) {
-			rsti(item)->pgrp_leader = gleader;
+		pid = pstree_pid_by_virt(item->pgid);
+		if (pid->state != TASK_UNDEF) {
+			BUG_ON(pid->state == TASK_THREAD);
+			rsti(item)->pgrp_leader = container_of(pid, struct pstree_item, pid);
 			continue;
 		}
 
@@ -636,9 +747,9 @@ static int prepare_pstree_ids(void)
 		if (current_pgid == item->pgid)
 			continue;
 
-		helper = alloc_pstree_helper();
-		if (helper == NULL)
-			return -1;
+		helper = container_of(pid, struct pstree_item, pid);
+		init_pstree_helper(helper);
+
 		helper->sid = item->sid;
 		helper->pgid = item->pgid;
 		helper->pid.virt = item->pgid;
@@ -689,8 +800,9 @@ static int prepare_pstree_kobj_ids(void)
 
 		if (!item->ids) {
 			if (item == root_item) {
-				cflags = opts.rst_namespaces_flags;
-				goto set_mask;
+				pr_err("No IDS for root task.\n");
+				pr_err("Images currupted or too old criu was used for dump.\n");
+				return -1;
 			}
 
 			continue;
@@ -730,7 +842,6 @@ static int prepare_pstree_kobj_ids(void)
 				return ret;
 		}
 
-set_mask:
 		rsti(item)->clone_flags = cflags;
 		if (parent)
 			/*
@@ -774,8 +885,41 @@ set_mask:
 int prepare_pstree(void)
 {
 	int ret;
+	pid_t pid_max = 0, kpid_max = 0;
+	int fd;
+	char buf[21];
+
+	fd = open_proc(PROC_GEN, PID_MAX_PATH);
+	if (fd >= 0) {
+		ret = read(fd, buf, sizeof(buf) - 1);
+		close(fd);
+		if (ret > 0) {
+			buf[ret] = 0;
+			kpid_max = strtoul(buf, NULL, 10);
+			pr_debug("kernel pid_max=%d\n", kpid_max);
+		}
+	}
+
+	ret = read_pstree_image(&pid_max);
+	pr_debug("pstree pid_max=%d\n", pid_max);
+
+	if (!ret && kpid_max && pid_max > kpid_max) {
+		/* Try to set kernel pid_max */
+		fd = open_proc_rw(PROC_GEN, PID_MAX_PATH);
+		if (fd == -1)
+			ret = -1;
+		else {
+			snprintf(buf, sizeof(buf), "%u", pid_max+1);
+			if (write(fd, buf, strlen(buf)) < 0) {
+				pr_perror("Can't set kernel pid_max=%s", buf);
+				ret = -1;
+			}
+			else
+				pr_info("kernel pid_max pushed to %s\n", buf);
+			close(fd);
+		}
+	}
 
-	ret = read_pstree_image();
 	if (!ret)
 		/*
 		 * Shell job may inherit sid/pgid from the current
@@ -810,13 +954,14 @@ bool restore_before_setsid(struct pstree_item *child)
 
 struct pstree_item *pstree_item_by_virt(pid_t virt)
 {
-	struct pstree_item *item;
+	struct pid *pid;
 
-	for_each_pstree_item(item) {
-		if (item->pid.virt == virt)
-			return item;
-	}
-	return NULL;
+	pid = pstree_pid_by_virt(virt);
+	if (pid == NULL)
+		return NULL;
+	BUG_ON(pid->state == TASK_THREAD);
+
+	return container_of(pid, struct pstree_item, pid);
 }
 
 struct pstree_item *pstree_item_by_real(pid_t real)
@@ -839,8 +984,3 @@ int pid_to_virt(pid_t real)
 		return item->pid.virt;
 	return 0;
 }
-
-bool pid_in_pstree(pid_t pid)
-{
-	return pstree_item_by_real(pid) != NULL;
-}
