@@ -17,7 +17,6 @@
 #include "unix_diag.h"
 #include "files.h"
 #include "file-ids.h"
-#include "image.h"
 #include "log.h"
 #include "util.h"
 #include "util-pie.h"
@@ -36,6 +35,24 @@
 #undef	LOG_PREFIX
 #define LOG_PREFIX "sk unix: "
 
+/*
+ * By-default, when dumping a unix socket, we should dump its peer
+ * as well. Which in turn means, we should dump the task(s) that have
+ * this peer opened.
+ *
+ * Sometimes, we can break this rule and dump only one end of the
+ * unix sockets pair, and on restore time connect() this end back to
+ * its peer.
+ *
+ * So, to resolve this situation we mark the peers we don't dump
+ * as "external" and require the --ext-unix-sk option.
+ */
+
+#define USK_EXTERN	(1 << 0)
+#define USK_SERVICE	(1 << 1)
+#define USK_CALLBACK	(1 << 2)
+#define USK_INHERIT	(1 << 3)
+
 typedef struct {
 	char			*dir;
 	unsigned int		udiag_vfs_dev;
@@ -55,6 +72,7 @@ struct unix_sk_desc {
 	unsigned int		nr_icons;
 	unsigned int		*icons;
 	unsigned char		shutdown;
+	bool			deleted;
 
 	mode_t			mode;
 	uid_t			uid;
@@ -186,10 +204,7 @@ static int resolve_rel_name(struct unix_sk_desc *sk, const struct fd_parms *p)
 	int mntns_root, i;
 	struct ns_id *ns;
 
-	for_each_pstree_item(task) {
-		if (task->pid.real == p->pid)
-			break;
-	}
+	task = pstree_item_by_real(p->pid);
 	if (!task) {
 		pr_err("Can't find task with pid %d\n", p->pid);
 		return -ENOENT;
@@ -323,6 +338,11 @@ static int dump_one_unix_fd(int lfd, u32 id, const struct fd_parms *p)
 		perms->gid	= userns_gid(sk->gid);
 	}
 
+	if (sk->deleted) {
+		ue->has_deleted = true;
+		ue->deleted	= sk->deleted;
+	}
+
 	sk_encode_shutdown(ue, sk->shutdown);
 
 	if (ue->peer) {
@@ -350,11 +370,7 @@ static int dump_one_unix_fd(int lfd, u32 id, const struct fd_parms *p)
 		 * until all sockets the program owns are processed.
 		 */
 		if (!peer->sd.already_dumped) {
-			if (list_empty(&peer->list)) {
-				show_one_unix("Add a peer", peer);
-				list_add_tail(&peer->list, &unix_sockets);
-			}
-
+			show_one_unix("Add a peer", peer);
 			list_add(&sk->peer_node, &peer->peer_list);
 			sk->fd = dup(lfd);
 			if (sk->fd < 0) {
@@ -371,12 +387,15 @@ static int dump_one_unix_fd(int lfd, u32 id, const struct fd_parms *p)
 				((ue->shutdown == SK_SHUTDOWN__BOTH)  &&
 				 (peer->shutdown != SK_SHUTDOWN__BOTH)) )) {
 			/*
-			 * On restore we assume, that stream pairs must
-			 * be shut down from one end only
+			 * Usually this doesn't happen, however it's possible if
+			 * socket was shut down before connect() (see sockets03.c test).
+			 * On restore we will shutdown both end (iow socktes will be in
+			 * matched state). This shoudn't be a problem, since kernel seems
+			 * to check both ends on read()/write(). Thus mismatched sockets behave
+			 * the same way as matched.
 			 */
-			pr_err("Shutdown mismatch %u:%d -> %u:%d\n",
+			pr_warn("Shutdown mismatch %u:%d -> %u:%d\n",
 					ue->ino, ue->shutdown, peer->sd.ino, peer->shutdown);
-			goto err;
 		}
 	} else if (ue->state == TCP_ESTABLISHED) {
 		const struct unix_sk_listen_icon *e;
@@ -447,7 +466,6 @@ dump:
 	if (list_empty(&sk->peer_node) && write_unix_entry(sk))
 		return -1;
 
-	list_del_init(&sk->list);
 	sk->sd.already_dumped = 1;
 
 	while (!list_empty(&sk->peer_list)) {
@@ -491,7 +509,7 @@ static int unix_process_name(struct unix_sk_desc *d, const struct unix_diag_msg
 
 	if (name[0] != '\0') {
 		struct unix_diag_vfs *uv;
-		bool drop_path = false;
+		bool deleted = false;
 		char rpath[PATH_MAX];
 		struct ns_id *ns;
 		struct stat st;
@@ -542,30 +560,21 @@ static int unix_process_name(struct unix_sk_desc *d, const struct unix_diag_msg
 
 			pr_info("unix: Dropping path %s for unlinked sk %#x\n",
 				name, m->udiag_ino);
-			drop_path = true;
+			deleted = true;
 		} else if ((st.st_ino != uv->udiag_vfs_ino) ||
 			   !phys_stat_dev_match(st.st_dev, uv->udiag_vfs_dev, ns, name)) {
 			pr_info("unix: Dropping path %s for unlinked bound "
 				"sk %#x.%#x real %#x.%#x\n",
 				name, (int)st.st_dev, (int)st.st_ino,
 				(int)uv->udiag_vfs_dev, (int)uv->udiag_vfs_ino);
-			drop_path = true;
-		}
-
-		if (drop_path) {
-			/*
-			 * When a socket is bound to unlinked file, we
-			 * just drop his name, since no one will access
-			 * it via one.
-			 */
-			xfree(name);
-			len = 0;
-			name = NULL;
+			deleted = true;
 		}
 
 		d->mode = st.st_mode;
 		d->uid	= st.st_uid;
 		d->gid	= st.st_gid;
+
+		d->deleted = deleted;
 	}
 
 postprone:
@@ -660,6 +669,7 @@ static int unix_collect_one(const struct unix_diag_msg *m,
 	}
 
 	sk_collect_one(m->udiag_ino, AF_UNIX, &d->sd);
+	list_add_tail(&d->list, &unix_sockets);
 	show_one_unix("Collected", d);
 
 	return 0;
@@ -740,9 +750,10 @@ int fix_external_unix_sockets(void)
 		FownEntry fown = FOWN_ENTRY__INIT;
 		SkOptsEntry skopts = SK_OPTS_ENTRY__INIT;
 
-		show_one_unix("Dumping extern", sk);
+		if (sk->sd.already_dumped)
+			continue;
 
-		BUG_ON(sk->sd.already_dumped);
+		show_one_unix("Dumping extern", sk);
 
 		fd_id_generate_special(NULL, &e.id);
 		e.ino		= sk->sd.ino;
@@ -944,31 +955,78 @@ static int bind_unix_sk(int sk, struct unix_sk_info *ui)
 	if (prep_unix_sk_cwd(ui, &cwd_fd))
 		return -1;
 
-	if (bind(sk, (struct sockaddr *)&addr,
-				sizeof(addr.sun_family) + ui->ue->name.len)) {
-		pr_perror("Can't bind socket");
-		goto done;
-	}
+	if (ui->ue->name.len) {
+		ret = bind(sk, (struct sockaddr *)&addr,
+				sizeof(addr.sun_family) + ui->ue->name.len);
+		if (ret < 0) {
+			if (ui->ue->has_deleted && ui->ue->deleted && errno == EADDRINUSE) {
+				char temp[PATH_MAX];
 
-	if (ui->ue->name.len && *ui->name && ui->ue->file_perms) {
-		FilePermsEntry *perms = ui->ue->file_perms;
-		char fname[PATH_MAX];
+				pr_info("found duplicate unix socket bound at %s\n", addr.sun_path);
 
-		if (ui->ue->name.len >= sizeof(fname)) {
-			pr_err("The file name is too long\n");
-			goto done;
+				ret = snprintf(temp, sizeof(temp), "%s-%s-%d", addr.sun_path, "criu-temp", getpid());
+				/* this shouldn't happen, since sun_addr is only 108 chars long */
+				if (ret < 0 || ret >= sizeof(temp)) {
+					pr_err("snprintf of %s failed?\n", addr.sun_path);
+					goto done;
+				}
+
+				ret = rename(addr.sun_path, temp);
+				if (ret < 0) {
+					pr_perror("couldn't move socket for binding");
+					goto done;
+				}
+
+				ret = bind(sk, (struct sockaddr *)&addr,
+						sizeof(addr.sun_family) + ui->ue->name.len);
+				if (ret < 0) {
+					pr_perror("Can't bind socket after move");
+					goto done;
+				}
+
+				ret = rename(temp, addr.sun_path);
+				if (ret < 0) {
+					pr_perror("couldn't move socket back");
+					goto done;
+				}
+
+				/* we've handled the deleted-ness of this
+				 * socket and we don't want to delete it later
+				 * since it's not /this/ socket.
+				 */
+				ui->ue->deleted = false;
+
+			} else {
+				pr_perror("Can't bind socket");
+				goto done;
+			}
 		}
 
-		memcpy(fname, ui->name, ui->ue->name.len);
-		fname[ui->ue->name.len] = '\0';
+		if (*ui->name && ui->ue->file_perms) {
+			FilePermsEntry *perms = ui->ue->file_perms;
+			char fname[PATH_MAX];
 
-		if (fchownat(AT_FDCWD, fname, perms->uid, perms->gid, 0) == -1) {
-			pr_perror("Unable to change file owner and group");
-			goto done;
+			if (ui->ue->name.len >= sizeof(fname)) {
+				pr_err("The file name is too long\n");
+				goto done;
+			}
+
+			memcpy(fname, ui->name, ui->ue->name.len);
+			fname[ui->ue->name.len] = '\0';
+
+			if (fchownat(AT_FDCWD, fname, perms->uid, perms->gid, 0) == -1) {
+				pr_perror("Unable to change file owner and group");
+				goto done;
+			}
+
+			if (fchmodat(AT_FDCWD, fname, perms->mode, 0) == -1) {
+				pr_perror("Unable to change file mode bits");
+				goto done;
+			}
 		}
 
-		if (fchmodat(AT_FDCWD, fname, perms->mode, 0) == -1) {
-			pr_perror("Unable to change file mode bits");
+		if (ui->ue->deleted && unlink((char *)ui->ue->name.data) < 0) {
+			pr_perror("failed to unlink %s\n", ui->ue->name.data);
 			goto done;
 		}
 	}
@@ -1022,11 +1080,7 @@ static int open_unixsk_pair_master(struct unix_sk_info *ui)
 	if (shutdown_unix_sk(sk[0], ui))
 		return -1;
 
-	tsk = socket(PF_UNIX, SOCK_DGRAM, 0);
-	if (tsk < 0) {
-		pr_perror("Can't make transport socket");
-		return -1;
-	}
+	tsk = get_service_fd(TRANSPORT_FD_OFF);
 
 	fle = file_master(&peer->d);
 	if (send_fd_to_peer(sk[1], fle, tsk)) {
@@ -1034,7 +1088,6 @@ static int open_unixsk_pair_master(struct unix_sk_info *ui)
 		return -1;
 	}
 
-	close(tsk);
 	close(sk[1]);
 
 	return sk[0];
@@ -1066,13 +1119,8 @@ static int open_unixsk_pair_slave(struct unix_sk_info *ui)
 	if (restore_socket_opts(sk, ui->ue->opts))
 		return -1;
 
-	if (ui->ue->type == SOCK_DGRAM)
-		/*
-		 * Stream socket's "slave" end will be shut down
-		 * together with master
-		 */
-		if (shutdown_unix_sk(sk, ui))
-			return -1;
+	if (shutdown_unix_sk(sk, ui))
+		return -1;
 
 	return sk;
 }
@@ -1154,7 +1202,7 @@ static int open_unixsk_standalone(struct unix_sk_info *ui)
 		 * The below is hack: we use that connect with AF_UNSPEC
 		 * clears socket's peer.
 		 */
-		if (connect(sk, &addr, sizeof(addr.sun_family))) {
+		if (connect(sk, (struct sockaddr *)&addr, sizeof(addr.sun_family))) {
 			pr_perror("Can't clear socket's peer");
 			return -1;
 		}
@@ -1330,10 +1378,29 @@ struct collect_image_info unix_sk_cinfo = {
 	.flags = COLLECT_SHARED,
 };
 
+static void interconnected_pair(struct unix_sk_info *ui, struct unix_sk_info *peer)
+{
+	struct fdinfo_list_entry *fle, *fle_peer;
+	/*
+	 * Select who will restore the pair. Check is identical to
+	 * the one in pipes.c and makes sure tasks wait for each other
+	 * in pids sorting order (ascending).
+	 */
+	fle = file_master(&ui->d);
+	fle_peer = file_master(&peer->d);
+
+	if (fdinfo_rst_prio(fle, fle_peer)) {
+		ui->flags |= USK_PAIR_MASTER;
+		peer->flags |= USK_PAIR_SLAVE;
+	} else {
+		peer->flags |= USK_PAIR_MASTER;
+		ui->flags |= USK_PAIR_SLAVE;
+	}
+}
+
 static int resolve_unix_peers(void *unused)
 {
 	struct unix_sk_info *ui, *peer;
-	struct fdinfo_list_entry *fle, *fle_peer;
 
 	list_for_each_entry(ui, &unix_sockets, list) {
 		if (ui->peer)
@@ -1358,25 +1425,10 @@ static int resolve_unix_peers(void *unused)
 		if (peer->ue->peer != ui->ue->ino)
 			continue;
 
-		/* socketpair or interconnected sockets */
 		peer->peer = ui;
 
-		/*
-		 * Select who will restore the pair. Check is identical to
-		 * the one in pipes.c and makes sure tasks wait for each other
-		 * in pids sorting order (ascending).
-		 */
-
-		fle = file_master(&ui->d);
-		fle_peer = file_master(&peer->d);
-
-		if (fdinfo_rst_prio(fle, fle_peer)) {
-			ui->flags |= USK_PAIR_MASTER;
-			peer->flags |= USK_PAIR_SLAVE;
-		} else {
-			peer->flags |= USK_PAIR_MASTER;
-			ui->flags |= USK_PAIR_SLAVE;
-		}
+		/* socketpair or interconnected sockets */
+		interconnected_pair(ui, peer);
 	}
 
 	pr_info("Unix sockets:\n");
