@@ -14,6 +14,7 @@
 #include <sys/sendfile.h>
 #include <fcntl.h>
 #include <poll.h>
+#include <arpa/inet.h>
 #include <sys/mount.h>
 #include <sys/param.h>
 #include <sys/types.h>
@@ -73,27 +74,27 @@ static char *xvstrcat(char *str, const char *fmt, va_list args)
 	delta = strlen(fmt) * 2;
 
 	do {
-		ret = -ENOMEM;
 		new = xrealloc(str, offset + delta);
-		if (new) {
-			va_copy(tmp, args);
-			ret = vsnprintf(new + offset, delta, fmt, tmp);
-			va_end(tmp);
-			if (ret >= delta) {
-				/* NOTE: vsnprintf returns the amount of bytes
-				 * to allocate. */
-				delta = ret +1;
-				str = new;
-				ret = 0;
-			}
+		if (!new) {
+			/* realloc failed. We must release former string */
+			xfree(str);
+			pr_err("Failed to allocate string\n");
+			return new;
 		}
-	} while (ret == 0);
 
-	if (ret == -ENOMEM) {
-		/* realloc failed. We must release former string */
-		pr_err("Failed to allocate string\n");
-		xfree(str);
-	} else if (ret < 0) {
+		va_copy(tmp, args);
+		ret = vsnprintf(new + offset, delta, fmt, tmp);
+		va_end(tmp);
+		if (ret < delta) /* an error, or all was written */
+			break;
+
+		/* NOTE: vsnprintf returns the amount of bytes
+		 * to allocate. */
+		delta = ret + 1;
+		str = new;
+	} while (1);
+
+	if (ret < 0) {
 		/* vsnprintf failed */
 		pr_err("Failed to print string\n");
 		xfree(new);
@@ -160,12 +161,13 @@ void pr_vma(unsigned int loglevel, const struct vma_area *vma_area)
 		return;
 
 	vma_opt_str(vma_area, opt);
-	print_on_level(loglevel, "%#"PRIx64"-%#"PRIx64" (%"PRIi64"K) prot %#x flags %#x st %#x off %#"PRIx64" "
+	print_on_level(loglevel, "%#"PRIx64"-%#"PRIx64" (%"PRIi64"K) prot %#x flags %#x fdflags %#o st %#x off %#"PRIx64" "
 			"%s shmid: %#"PRIx64"\n",
 			vma_area->e->start, vma_area->e->end,
 			KBYTES(vma_area_len(vma_area)),
 			vma_area->e->prot,
 			vma_area->e->flags,
+			vma_area->e->fdflags,
 			vma_area->e->status,
 			vma_area->e->pgoff,
 			opt, vma_area->e->shmid);
@@ -217,7 +219,7 @@ int reopen_fd_as_safe(char *file, int line, int new_fd, int old_fd, bool allow_r
 	return 0;
 }
 
-int move_img_fd(int *img_fd, int want_fd)
+int move_fd_from(int *img_fd, int want_fd)
 {
 	if (*img_fd == want_fd) {
 		int tmp;
@@ -385,14 +387,14 @@ static int service_fd_id = 0;
 
 int init_service_fd(void)
 {
-	struct rlimit rlimit;
+	struct rlimit64 rlimit;
 
 	/*
 	 * Service FDs are those that most likely won't
 	 * conflict with any 'real-life' ones
 	 */
 
-	if (getrlimit(RLIMIT_NOFILE, &rlimit)) {
+	if (syscall(__NR_prlimit64, getpid(), RLIMIT_NOFILE, NULL, &rlimit)) {
 		pr_perror("Can't get rlimit");
 		return -1;
 	}
@@ -408,6 +410,11 @@ static int __get_service_fd(enum sfd_type type, int service_fd_id)
 	return service_fd_rlim_cur - type - SERVICE_FD_MAX * service_fd_id;
 }
 
+int service_fd_min_fd(void)
+{
+	return service_fd_rlim_cur - (SERVICE_FD_MAX - 1) - SERVICE_FD_MAX * service_fd_id;
+}
+
 static DECLARE_BITMAP(sfd_map, SERVICE_FD_MAX);
 
 int reserve_service_fd(enum sfd_type type)
@@ -557,17 +564,6 @@ int is_anon_link_type(char *link, char *type)
 	return !strcmp(link, aux);
 }
 
-void *shmalloc(size_t bytes)
-{
-	return rst_mem_alloc(bytes, RM_SHARED);
-}
-
-/* Only last chunk can be released */
-void shfree_last(void *ptr)
-{
-	rst_mem_free_last(RM_SHARED);
-}
-
 #define DUP_SAFE(fd, out)						\
 	({							\
 		int ret__;					\
@@ -617,9 +613,9 @@ int cr_system_userns(int in, int out, int err, char *cmd,
 		}
 
 		if (out < 0)
-			out = log_get_fd();
+			out = DUP_SAFE(log_get_fd(), out_chld);
 		if (err < 0)
-			err = log_get_fd();
+			err = DUP_SAFE(log_get_fd(), out_chld);
 
 		/*
 		 * out, err, in should be a separate fds,
@@ -631,8 +627,8 @@ int cr_system_userns(int in, int out, int err, char *cmd,
 		if (out == in)
 			out = DUP_SAFE(out, out_chld);
 
-		if (move_img_fd(&out, STDIN_FILENO) ||
-		    move_img_fd(&err, STDIN_FILENO))
+		if (move_fd_from(&out, STDIN_FILENO) ||
+		    move_fd_from(&err, STDIN_FILENO))
 			goto out_chld;
 
 		if (in < 0) {
@@ -642,7 +638,7 @@ int cr_system_userns(int in, int out, int err, char *cmd,
 				goto out_chld;
 		}
 
-		if (move_img_fd(&err, STDOUT_FILENO))
+		if (move_fd_from(&err, STDOUT_FILENO))
 			goto out_chld;
 
 		if (reopen_fd_as_nocheck(STDOUT_FILENO, out))
@@ -670,7 +666,8 @@ out_chld:
 				pr_err("exited, status=%d\n", WEXITSTATUS(status));
 			break;
 		} else if (WIFSIGNALED(status)) {
-			pr_err("killed by signal %d\n", WTERMSIG(status));
+			pr_err("killed by signal %d: %s\n", WTERMSIG(status),
+				strsignal(WTERMSIG(status)));
 			break;
 		} else if (WIFSTOPPED(status)) {
 			pr_err("stopped by signal %d\n", WSTOPSIG(status));
@@ -797,7 +794,6 @@ struct vma_area *alloc_vma_area(void)
 	if (p) {
 		p->e = (VmaEntry *)(p + 1);
 		vma_entry__init(p->e);
-		p->vm_file_fd = -1;
 		p->e->fd = -1;
 	}
 
@@ -941,21 +937,6 @@ int fd_has_data(int lfd)
 	return ret;
 }
 
-size_t read_into_buffer(int fd, char *buff, size_t size)
-{
-	size_t n = 0;
-	size_t curr = 0;
-
-	while (1) {
-		n  = read(fd, buff + curr, size - curr);
-		if (n < 1)
-			return n;
-		curr += n;
-		if (curr == size)
-			return size;
-	}
-}
-
 int make_yard(char *path)
 {
 	if (mount("none", path, "tmpfs", 0, NULL)) {
@@ -1074,3 +1055,132 @@ void print_data(unsigned long addr, unsigned char *data, size_t size)
 		pr_msg("|\n");
 	}
 }
+
+static int get_sockaddr_in(struct sockaddr_in *addr, char *host)
+{
+	memset(addr, 0, sizeof(*addr));
+	addr->sin_family = AF_INET;
+
+	if (!host)
+		addr->sin_addr.s_addr = INADDR_ANY;
+	else if (!inet_aton(host, &addr->sin_addr)) {
+		pr_perror("Bad server address");
+		return -1;
+	}
+
+	addr->sin_port = opts.port;
+	return 0;
+}
+
+int setup_tcp_server(char *type)
+{
+	int sk = -1;
+	struct sockaddr_in saddr;
+	socklen_t slen = sizeof(saddr);
+
+	pr_info("Starting %s server on port %u\n", type, (int)ntohs(opts.port));
+
+	sk = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+	if (sk < 0) {
+		pr_perror("Can't init %s server", type);
+		return -1;
+	}
+
+	if (get_sockaddr_in(&saddr, opts.addr))
+		goto out;
+
+	if (bind(sk, (struct sockaddr *)&saddr, slen)) {
+		pr_perror("Can't bind %s server", type);
+		goto out;
+	}
+
+	if (listen(sk, 1)) {
+		pr_perror("Can't listen on %s server socket", type);
+		goto out;
+	}
+
+	/* Get socket port in case of autobind */
+	if (opts.port == 0) {
+		if (getsockname(sk, (struct sockaddr *)&saddr, &slen)) {
+			pr_perror("Can't get %s server name", type);
+			goto out;
+		}
+
+		opts.port = ntohs(saddr.sin_port);
+		pr_info("Using %u port\n", opts.port);
+	}
+
+	return sk;
+out:
+	close(sk);
+	return -1;
+}
+
+int run_tcp_server(bool daemon_mode, int *ask, int cfd, int sk)
+{
+	int ret;
+	struct sockaddr_in caddr;
+	socklen_t clen = sizeof(caddr);
+
+	if (daemon_mode) {
+		ret = cr_daemon(1, 0, ask, cfd);
+		if (ret == -1) {
+			pr_err("Can't run in the background\n");
+			goto out;
+		}
+		if (ret > 0) { /* parent task, daemon started */
+			close_safe(&sk);
+			if (opts.pidfile) {
+				if (write_pidfile(ret) == -1) {
+					pr_perror("Can't write pidfile");
+					kill(ret, SIGKILL);
+					waitpid(ret, NULL, 0);
+					return -1;
+				}
+			}
+
+			return ret;
+		}
+	}
+
+	if (sk >= 0) {
+		ret = *ask = accept(sk, (struct sockaddr *)&caddr, &clen);
+		if (*ask < 0)
+			pr_perror("Can't accept connection to server");
+		else
+			pr_info("Accepted connection from %s:%u\n",
+					inet_ntoa(caddr.sin_addr),
+					(int)ntohs(caddr.sin_port));
+		close(sk);
+	}
+
+	return 0;
+out:
+	close(sk);
+	return -1;
+}
+
+int setup_tcp_client(char *addr)
+{
+	struct sockaddr_in saddr;
+	int sk;
+
+	pr_info("Connecting to server %s:%u\n", addr, (int)ntohs(opts.port));
+
+	if (get_sockaddr_in(&saddr, addr))
+		return -1;
+
+	sk = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+	if (sk < 0) {
+		pr_perror("Can't create socket");
+		return -1;
+	}
+
+	if (connect(sk, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
+		pr_perror("Can't connect to server");
+		close(sk);
+		return -1;
+	}
+
+	return sk;
+}
